#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


"""
Utilities for working with data for an Ax experiment.
Unlike the Ax Data object, "data" here refers to both the arm
parameterizations and the metric observations (from a Data object).
"""

from __future__ import annotations

from collections.abc import Iterable
from copy import deepcopy
from dataclasses import dataclass
from typing import Any

import numpy as np
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.map_data import MAP_KEY, MapData
from ax.core.map_metric import MapMetric
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.trial_status import NON_ABANDONED_STATUSES, TrialStatus
from ax.core.types import TParameterization
from ax.exceptions.core import UnsupportedError
from ax.utils.common.constants import Keys
from pandas import DataFrame, MultiIndex, Series
from pyre_extensions import none_throws


@dataclass(frozen=True)
class DataLoaderConfig:
    """This dataclass contains parameters that control the behavior
    of `Adapter._set_and_filter_training_data`.

    Args:
        fit_out_of_design: If specified, all training data are used.
            Otherwise, only in design points are used. Note that in-design-ness is
            determined after expanding the modeling space, if applicable.
        fit_abandoned: Whether data for abandoned arms or trials should be included in
            model training data. If `False`, only non-abandoned points are returned.
        fit_only_completed_map_metrics: Whether to fit a model to map metrics only when
            the trial is completed. This is useful for applications like modeling
            partially completed learning curves in AutoML.
        latest_rows_per_group: If specified and data is an instance of MapData, uses
            MapData.latest() with `latest_rows_per_group` to retrieve the most recent
            rows for each group. Useful in cases where learning curves are frequently
            updated, preventing an excessive number of Observation objects.
        limit_rows_per_metric: Subsample the map data so that the total number of
            rows per metric is limited by this value.
        limit_rows_per_group: Subsample the map data so that the number of rows
            in the `MAP_KEY` column for each (arm, metric) is limited by this value.
    """

    fit_out_of_design: bool = False
    fit_abandoned: bool = False
    fit_only_completed_map_metrics: bool = False
    latest_rows_per_group: int | None = 1
    limit_rows_per_metric: int | None = None
    limit_rows_per_group: int | None = None

    def __post_init__(self) -> None:
        if self.latest_rows_per_group is not None and (
            self.limit_rows_per_metric is not None
            or self.limit_rows_per_group is not None
        ):
            raise UnsupportedError(
                "`latest_rows_per_group` must be None if either of "
                "`limit_rows_per_metric` or `limit_rows_per_group` is specified."
            )

    @property
    def statuses_to_fit(self) -> set[TrialStatus]:
        """The data from trials in these statuses will be used to fit the model
        for non map metrics. Defaults to all trial statuses if
        `fit_abandoned is True` and all statuses except ABANDONED, otherwise.
        """
        if self.fit_abandoned:
            return set(TrialStatus)
        return NON_ABANDONED_STATUSES

    @property
    def statuses_to_fit_map_metric(self) -> set[TrialStatus]:
        """The data from trials in these statuses will be used to fit the model
        for map metrics. Defaults to only COMPLETED trials if
        `fit_only_completed_map_metrics is True` and to `statuses_to_fit`, otherwise.
        """
        if self.fit_only_completed_map_metrics:
            return {TrialStatus.COMPLETED}
        return self.statuses_to_fit


@dataclass
class ExperimentData:
    """A container for the data from arms and observations of an experiment.

    This is intended as a lightweight container for the data that gets processed
    within the ``Adapter`` / ``Transform`` layer. This will be passed through
    transforms, allowing for column-wise, vectorized transforming of the values.
    This can then be used to construct any necessary data structures for the
    ``Generator``, such as the ``SupervisedDataset`` objects constructed
    in ``TorchAdapter``.

    NOTE: This should only be constructed using the ``extract_experiment_data`` helper.
    Construction of the underlying dataframes manually from scratch can be quite
    complicated and error prone due to the multi-index involved in both the
    index and columns of the dataframes.

    Attributes:
        arm_data: A dataframe, indexed by (trial_index, arm_name), containing the
            the parameterization of each arm, with one column per parameter, and
            a column for the metadata. Each row corresponds to the parameterization
            and metadata for the given (trial_index, arm_name) pair.
        observation_data: A dataframe, indexed by (trial_index, arm_name[, *map_keys])
            map_keys being optional, containing the mean and sem observations for each
            metric. The columns of the dataframe are multi-indexed, with the first level
            being "mean" or "sem" and the second level being the metric name.
            This is typically constructed by pivoting `(Map)Data.true_df`.
            If the `Data` object contains additional metadata columns like `start_time`
            and `end_time`, these will be carried onto `observation_data`. The metadata
            columns will be indexed with the first level being "metadata" and the second
            level keeping the original column name.

    Example with non-map data:
        >>> experiment = Experiment(...)  # An experiment with non-map Data.
        >>> experiment_data = extract_experiment_data(
        ...     experiment=experiment,
        ...     data_loader_config=DataLoaderConfig(),
        ... )
        >>> print(experiment_data.arm_data)
                                     x         y
        trial_index arm_name
        0           0_0       0.539644  0.878898
        1           1_0       0.341846  0.213774
        3           3_0       0.837545  0.732969
        >>> print(experiment_data.observation_data)
                             mean      sem
        metric_name            m1   m2  m1  m2
        trial_index arm_name
        0           0_0       0.1  1.0 NaN NaN
        1           1_0       0.2  2.0 NaN NaN

    Example with map data:
        >>> experiment = Experiment(...)  # An experiment with MapData.
        >>> experiment_data = extract_experiment_data(
        ...     experiment=experiment,
        ...     data_loader_config=DataLoaderConfig(
        ...         fit_only_completed_map_metrics=False,
        ...         latest_rows_per_group=None,
        ...     ),
        ... )
        >>> print(experiment_data.arm_data)
                               x1   x2
        trial_index arm_name
        0           0_0       0.0  0.0
        1           1_0       1.0  1.0
        >>> print(experiment_data.observation_data)
                                            mean               sem
        metric_name                        branin branin_map branin branin_map
        trial_index arm_name timestamp
        0           0_0      0.0        55.602113  55.602113    0.0        0.0
                             1.0              NaN  55.602113    NaN        0.0
                             2.0              NaN  55.602113    NaN        0.0
                             3.0              NaN  55.602113    NaN        0.0
        1           1_0      0.0        27.702906  27.702906    0.0        0.0
                             1.0              NaN  27.702906    NaN        0.0
    """

    arm_data: DataFrame
    observation_data: DataFrame

    def filter_by_arm_names(self, arm_names: Iterable[str]) -> ExperimentData:
        """
        Returns a new ``ExperimentData`` object that is filtered to only include
        the rows corresponding to arms in ``arm_names``.
        """
        return ExperimentData(
            arm_data=self.arm_data[
                self.arm_data.index.get_level_values("arm_name").isin(arm_names)
            ],
            observation_data=self.observation_data[
                self.observation_data.index.get_level_values("arm_name").isin(arm_names)
            ],
        )

    def filter_by_trial_index(self, trial_indices: Iterable[int]) -> ExperimentData:
        """
        Returns a new ``ExperimentData`` object that is filtered to only include
        the rows corresponding to trials in ``trial_indices``.
        """
        return ExperimentData(
            arm_data=self.arm_data[
                self.arm_data.index.get_level_values("trial_index").isin(trial_indices)
            ],
            observation_data=self.observation_data[
                self.observation_data.index.get_level_values("trial_index").isin(
                    trial_indices
                )
            ],
        )

    def filter_latest_observations(self) -> ExperimentData:
        """
        Returns a new ``ExperimentData`` object, where the ``observation_data`` is
        filtered to only include the latest (according to map key index) value
        for each ``(trial_index, arm_name)`` pair. The resulting ``observation_data``
        will not have the map keys in its index.

        NOTE: This is an expensive operation. Use sparingly!
        """
        if len(self.observation_data.index.names) == 2:
            # No map key in the observation data. Nothing to filter.
            return deepcopy(self)
        if len(self.observation_data.index.names) != 3:
            raise UnsupportedError(
                "Filtering latest observations is not supported when the index "
                f"includes multiple map keys. Got {self.observation_data.index=}"
            )
        # This sorts each (trial_index, arm_name) group by the map key value,
        # fills NaNs (missing observation for a given progression) in each column
        # with the latest non-null value, and gets the last
        # row, which is now filled with all the latest observations.
        # The resulting dataframe has the highest progression observation for each
        # (trial_index, arm_name) pair for each metric.
        # TODO: See if we can do this more efficiently, maybe by avoiding groupby.apply.
        observation_data = (
            self.observation_data.groupby(
                level=["trial_index", "arm_name"], group_keys=False
            )
            # With map keys, we expect this to be pre-sorted but we can't guarantee.
            .apply(lambda df: df.sort_index(level=2).ffill().tail(1))
            .droplevel(2)  # Remove map key from the index.
        )
        return ExperimentData(
            arm_data=self.arm_data.copy(),
            observation_data=observation_data,
        )

    def convert_to_list_of_observations(self) -> list[Observation]:
        """Converts the ``ExperimentData`` to a list of ``Observation`` objects.

        This is useful for compatibility with some older methods that expect
        the `Adapter` training data to be a list of ``Observation`` objects.
        """
        has_map_keys = self.observation_data.index.nlevels > 2
        observations = []
        # pyre-ignore [23]: Pyre doesn't know the structure of the index.
        for (trial_index, arm_name), row in self.arm_data.iterrows():
            obs_ft_base = ObservationFeatures(
                # NOTE: It is crucial to pop metadata first here.
                # Otherwise, it'd end up in parameters.
                metadata=row.pop("metadata"),
                parameters=row.dropna().to_dict(),
                trial_index=trial_index,
            )
            # Different indexing in the two cases to ensure we get a dataframe.
            ind = (trial_index, arm_name)
            ind = ind if has_map_keys else [ind]
            data_rows = self.observation_data.loc[ind]
            has_multiple_rows = len(data_rows) > 1
            # Looping here to ensure we can capture different progression values
            # as different `Observation` objects.
            for idx in data_rows.index:
                # Keeping it as a df for consistent indexing.
                row_df = data_rows.loc[[idx]]
                # Only include metrics that have data.
                metric_names = list(row_df["mean"].dropna(axis="columns").columns)
                if len(metric_names) == 0:
                    continue
                if has_multiple_rows:
                    obs_ft = obs_ft_base.clone()
                else:
                    # No need to clone if there is only one row.
                    obs_ft = obs_ft_base
                if has_map_keys:
                    # Add map key to metadata as expected in ObservationFeatures.
                    none_throws(obs_ft.metadata)[data_rows.index.name] = idx
                obs_data = ObservationData(
                    metric_names=metric_names,
                    means=row_df["mean"][metric_names].to_numpy().reshape(-1),
                    covariance=np.diag(
                        np.square(row_df["sem"][metric_names].to_numpy().reshape(-1))
                    ),
                )
                observations.append(
                    Observation(features=obs_ft, data=obs_data, arm_name=arm_name)
                )
        return observations

    @property
    def metric_names(self) -> list[str]:
        """The list of metric names that are available on ``observation_data``."""
        try:
            return list(self.observation_data["mean"].columns)
        except KeyError:
            # No data is available, return empty list.
            return []

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ExperimentData):
            return False
        return self.arm_data.equals(other.arm_data) and self.observation_data.equals(
            other.observation_data
        )


def extract_experiment_data(
    experiment: Experiment,
    data_loader_config: DataLoaderConfig,
    data: Data | None = None,
) -> ExperimentData:
    """Extract ``ExperimentData`` from the trials & ``Data`` of an experiment.

    Args:
        experiment: The experiment to extract data from.
        data_loader_config: A DataLoaderConfig of options for loading data. See the
            docstring of DataLoaderConfig for more details.
        data: The observation data for the metrics of the experiment.
            Optional, defaults to `experiment.lookup_data()`.
    """
    data = data or experiment.lookup_data()
    arm_data = _extract_arm_data(experiment=experiment)
    observation_data = _extract_observation_data(
        experiment=experiment, data_loader_config=data_loader_config, data=data
    )
    # Filter arm_data to only include the rows in observation_data.
    index = observation_data.index
    if (num_levels := len(index.names)) > 2:
        # Keep only the first two levels: trial_index, arm_name.
        index = index.droplevel(level=list(range(2, num_levels)))
    arm_data = arm_data.loc[arm_data.index.isin(index)]
    return ExperimentData(arm_data=arm_data, observation_data=observation_data)


def _extract_arm_data(experiment: Experiment) -> DataFrame:
    """Extract a dataframe containing the trial index, arm name,
    parameterizations, and metadata from the given experiment.

    The dataframe will include a row for each (trial_index, arm_name) pair,
    as long as the corresponding trial has some data attached to the experiment.

    Args:
        experiment: The experiment to extract arms from.
    """
    records: dict[tuple[int, str], TParameterization] = {}
    for trial_index, trial in experiment.trials.items():
        for arm in trial.arms:
            column_values: dict[str, Any] = arm.parameters
            metadata = trial._get_candidate_metadata(arm.name) or {}
            if Keys.TRIAL_COMPLETION_TIMESTAMP not in metadata:
                if trial._time_completed is not None:
                    metadata[Keys.TRIAL_COMPLETION_TIMESTAMP] = (
                        trial._time_completed
                    ).timestamp()
            column_values["metadata"] = metadata
            records[(trial_index, arm.name)] = column_values
    if records:
        df = DataFrame.from_dict(records, orient="index")
        df.index.names = ["trial_index", "arm_name"]
    else:
        # No data, return an empty dataframe with the correct index & columns.
        index = MultiIndex.from_tuples([], names=["trial_index", "arm_name"])
        df = DataFrame(index=index, columns=list(experiment.parameters) + ["metadata"])
    return df


def _extract_observation_data(
    experiment: Experiment,
    data_loader_config: DataLoaderConfig,
    data: Data | None = None,
) -> DataFrame:
    """Extracts a dataframe containing filtered observations for the metrics
    on the experiment.

    See `extract_experiment_data` for a description of the arguments.

    Returns:
        A dataframe filtered to only include observations from the given statuses
        to include, and pivoted to be indexed by (trial_index, arm_name, *map_keys)
        and to have columns "mean" & "sem" for each metric.
        If `data` contains additional metadata columns like `start_time` and `end_time`
        they will be added to the pivoted dataframe as additional columns. The columns
        are labeled hierarchically (with a ``pd.MultiIndex`` structure) where the top
        level is "metadata" and the lower level keeps the original column name.
        For example, if the original DataFrame from ``Data`` includes a column
        ``start_time``, then ``observation_data`` will have a column that can be
        retrived with ``("metadata", "start_time")``.
    """
    data = data if data is not None else experiment.lookup_data()
    if isinstance(data, MapData):
        if data_loader_config.latest_rows_per_group is not None:
            data = data.latest(
                rows_per_group=data_loader_config.latest_rows_per_group,
            )
        elif (
            data_loader_config.limit_rows_per_metric is not None
            or data_loader_config.limit_rows_per_group is not None
        ):
            data = data.subsample(
                limit_rows_per_metric=data_loader_config.limit_rows_per_metric,
                limit_rows_per_group=data_loader_config.limit_rows_per_group,
                include_first_last=True,
            )

    df = data.true_df
    # Filter out rows for invalid statuses.
    to_keep = Series(index=df.index, data=False)
    trial_statuses = df["trial_index"].map(
        {trial_index: trial.status for trial_index, trial in experiment.trials.items()}
    )
    # If there are abandoned arms, mark the corresponding rows as abandoned.
    abandoned_arms = {
        i: {arm.name for arm in trial.abandoned_arms}
        for i, trial in experiment.trials.items()
        if trial.abandoned_arms
    }
    if abandoned_arms:
        is_abandoned = df[["trial_index", "arm_name"]].apply(
            lambda row: row["trial_index"] in abandoned_arms
            and row["arm_name"] in abandoned_arms[row["trial_index"]],
            axis=1,
        )
        trial_statuses[is_abandoned] = TrialStatus.ABANDONED
    # Check against valid statuses and filter the rows.
    for metric in experiment.metrics.values():
        valid_statuses = (
            data_loader_config.statuses_to_fit_map_metric
            if isinstance(metric, MapMetric) and metric.has_map_data
            else data_loader_config.statuses_to_fit
        )
        to_keep |= (df["metric_name"] == metric.name) & trial_statuses.isin(
            valid_statuses
        )
    df = df.loc[to_keep]
    # If df is empty, add mean & sem columns to facilitate pivoting.
    if df.empty:
        df = df.assign(mean=None, sem=None)

    # Identify potential metadata columns.
    index_cols = ["trial_index", "arm_name"]
    if isinstance(data, MapData):
        index_cols.append(MAP_KEY)

    standard_columns = set(index_cols).union({"metric_name", "mean", "sem"})
    metadata_columns = [col for col in df.columns if col not in standard_columns]

    # Pivot the df to be indexed by (trial_index, arm_name, *map_key)
    # and to have columns "mean" & "sem" for each metric.
    observation_data = df.pivot(
        columns="metric_name", index=index_cols, values=["mean", "sem"]
    )

    # If metadata columns exist, add them to the pivoted dataframe.
    if metadata_columns:
        # Create a dataframe with just the index columns and metadata columns.
        metadata_df = df[index_cols + metadata_columns]
        # Set the index to match the observation_data index.
        # All null rows are dropped to still capture the metadata in the next step
        # if it exists for only a subset of metrics.
        metadata_df = metadata_df.set_index(index_cols).dropna(how="all")
        # Drop duplicates to ensure we have only one row per unique index.
        # This is necessary when there are multiple metrics.
        metadata_df = metadata_df.loc[~metadata_df.index.duplicated(keep="first")]
        # Create a multi-index for the columns to facilitate the merge.
        metadata_df.columns = MultiIndex.from_product([["metadata"], metadata_columns])
        # Join to the main dataframe.
        observation_data = observation_data.join(metadata_df)

    return observation_data



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardGroup
from ax.analysis.plotly.marginal_effects import MarginalEffectsPlot
from ax.analysis.plotly.top_surfaces import TopSurfacesAnalysis
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.utils import is_bandit_experiment
from ax.exceptions.core import DataRequiredError, UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import override


INSIGHTS_CARDGROUP_TITLE = "Insights Analysis"

INSIGHTS_CARDGROUP_SUBTITLE = (
    "Insight Analyses display information to help understand the underlying "
    "experiment i.e parameter and metric relationships learned by the Ax model."
    "Use this information to better understand your experiment space and users."
)


class InsightsAnalysis(Analysis):
    """
    An Analysis that provides insights into the optimization process.

    For continuous and mixed search spaces, this includes sensitivity plots,
    slice plots, and contour plots.

    For bandit experiments, it returns a bandit rollout plot instead.
    """

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardGroup:
        if experiment is None:
            raise UserInputError("InsightsAnalysis requires an Experiment.")

        if experiment.lookup_data().df.empty:
            raise DataRequiredError(
                "Cannot compute InsightsAnalysis, Experiment has no data."
            )

        # Relativize the effects if the status quo is set and there are BatchTrials
        # present.
        relativize = experiment.status_quo is not None and any(
            isinstance(trial, BatchTrial) for trial in experiment.trials.values()
        )

        # If the Experiment has an OptimizationConfig set, extract the objective and
        # constraint names.
        objective_names = []
        constraint_names = []
        if (optimization_config := experiment.optimization_config) is not None:
            objective_names = optimization_config.objective.metric_names
            constraint_names = [
                constraint.metric.name
                for constraint in optimization_config.outcome_constraints
            ]

        is_bandit = generation_strategy and is_bandit_experiment(
            generation_strategy_name=generation_strategy.name
        )

        # For each objective and constraint, compute a sensitivity analysis and plot
        # the top 3 surfaces. Collect the bar (sensitivity) plots, slice plots, and
        # contour plots in separate lists.
        top_surfaces_groups = (
            [
                TopSurfacesAnalysis(
                    metric_name=metric_name,
                    top_k=3,
                    relativize=relativize,
                ).compute_or_error_card(
                    experiment=experiment,
                    generation_strategy=generation_strategy,
                    adapter=adapter,
                )
                for metric_name in [*objective_names, *constraint_names]
            ]
            if not is_bandit
            else []
        )

        # Add MarginalEffectsPlot for bandit experiments
        marginal_effects_groups = (
            [
                MarginalEffectsPlot(metric_name=metric_name).compute_or_error_card(
                    experiment=experiment,
                    generation_strategy=generation_strategy,
                    adapter=adapter,
                )
                for metric_name in [*objective_names, *constraint_names]
            ]
            if is_bandit
            else []
        )

        return self._create_analysis_card_group(
            title=INSIGHTS_CARDGROUP_TITLE,
            subtitle=INSIGHTS_CARDGROUP_SUBTITLE,
            children=[*top_surfaces_groups, *marginal_effects_groups],
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
References

.. [Papenmeier2025hd]
    L. Papenmeier, M. Poloczek, and L. Nardi. Understanding High-Dimensional Bayesian
        Optimization. Arxiv, 2025.
"""

from __future__ import annotations

from math import log, sqrt
from typing import Any, Sequence

import torch
from ax.exceptions.core import AxError
from gpytorch.constraints import Interval
from gpytorch.constraints.constraints import GreaterThan
from gpytorch.kernels import PeriodicKernel
from gpytorch.kernels.matern_kernel import MaternKernel
from gpytorch.kernels.rbf_kernel import RBFKernel
from gpytorch.kernels.scale_kernel import ScaleKernel
from gpytorch.priors.torch_priors import LogNormalPrior, Prior


class ScaleMaternKernel(ScaleKernel):
    def __init__(
        self,
        ard_num_dims: int | None = None,
        batch_shape: torch.Size | None = None,
        lengthscale_prior: Prior | None = None,
        outputscale_prior: Prior | None = None,
        lengthscale_constraint: Interval | None = None,
        outputscale_constraint: Interval | None = None,
        **kwargs: Any,
    ) -> None:
        r"""
        Args:
            ard_num_dims: The number of lengthscales.
            batch_shape: The batch shape.
            lengthscale_prior: The prior over the lengthscale parameter.
            outputscale_prior: The prior over the scaling parameter.
            lengthscale_constraint: Optionally provide a lengthscale constraint.
            outputscale_constraint: Optionally provide a output scale constraint.

        Returns: None
        """
        base_kernel = MaternKernel(
            nu=2.5,
            ard_num_dims=ard_num_dims,
            lengthscale_constraint=lengthscale_constraint,
            lengthscale_prior=lengthscale_prior,
            batch_shape=batch_shape,
        )
        super().__init__(
            base_kernel=base_kernel,
            outputscale_prior=outputscale_prior,
            outputscale_constraint=outputscale_constraint,
            **kwargs,
        )


class TemporalKernel(ScaleKernel):
    """A product kernel of a periodic kernel and a Matern kernel.

    The periodic kernel computes the similarity between temporal
    features such as the time of day.

    The Matern kernel computes the similarity between the tunable
    parameters.
    """

    def __init__(
        self,
        dim: int,
        temporal_features: list[int],
        matern_ard_num_dims: int | None = None,
        batch_shape: torch.Size | None = None,
        lengthscale_prior: Prior | None = None,
        temporal_lengthscale_prior: Prior | None = None,
        period_length_prior: Prior | None = None,
        fixed_period_length: float | None = None,
        outputscale_prior: Prior | None = None,
        lengthscale_constraint: Interval | None = None,
        outputscale_constraint: Interval | None = None,
        temporal_lengthscale_constraint: Interval | None = None,
        period_length_constraint: Interval | None = None,
        **kwargs: Any,
    ) -> None:
        r"""
        Args:
            dim: The input dimension.
            temporal_features: The features to pass to the periodic kernel.
            matern_ard_num_dims: The number of lengthscales. This must be
                equal to the total number of parameters (excluding
                temporal parameters)
            batch_shape: The batch shape.
            lengthscale_prior: The prior over the lengthscale parameters.
            temporal_lengthscale_prior: The prior over the lengthscale parameters
                for the periodic kernel.
            period_length_prior: The prior over the period length.
            fixed_period_length: A fixed period length for the periodic kernel.
                If provided, the period length will not be tuned with the other
                hyperparameters.
            outputscale_prior: The prior over the scaling parameter.
            lengthscale_constraint: Optionally provide a lengthscale constraint.
            outputscale_constraint: Optionally provide a output scale constraint.
            temporal_lengthscale_constraint: Optionally provide a
            lengthscale constraint for the periodic kernel.period_length_constraint:
                Optionally provide a constraint for the period length.

        """
        if len(temporal_features) == 0:
            raise AxError(
                "The temporal kernel should only be used if there "
                "are temporal features."
            )
        if fixed_period_length is not None and (
            period_length_prior is not None or period_length_constraint is not None
        ):
            raise ValueError(
                "If `fixed_period_length` is provided, then `period_length_prior` "
                "and `period_length_constraint` are not used."
            )
        non_temporal_dims = sorted(set(range(dim)) - set(temporal_features))
        matern_kernel = MaternKernel(
            nu=2.5,
            ard_num_dims=matern_ard_num_dims,
            lengthscale_prior=lengthscale_prior,
            active_dims=non_temporal_dims,
            batch_shape=batch_shape,
            lengthscale_constraint=lengthscale_constraint,
        )
        periodic_kernel = PeriodicKernel(
            ard_num_dims=len(temporal_features),
            active_dims=temporal_features,
            lengthscale_prior=temporal_lengthscale_prior,
            period_length_prior=period_length_prior,
            lengthscale_constraint=temporal_lengthscale_constraint,
            period_length_constraint=period_length_constraint,
            batch_shape=batch_shape,
        )
        if fixed_period_length is not None:
            periodic_kernel.raw_period_length.requires_grad_(False)
            periodic_kernel.period_length = fixed_period_length
        super().__init__(
            base_kernel=matern_kernel * periodic_kernel,
            outputscale_prior=outputscale_prior,
            outputscale_constraint=outputscale_constraint,
            **kwargs,
        )


def get_lengthscale_prior_and_initial_value(
    ard_num_dims: int | None, mle: bool
) -> tuple[LogNormalPrior | None, float]:
    if ard_num_dims is None:
        ard_num_dims = 1
    if mle:
        # initial value comes from [Papenmeier2025hd]_.
        return None, sqrt(ard_num_dims) / 10
    lengthscale_prior = LogNormalPrior(
        loc=sqrt(2) + log(ard_num_dims) * 0.5, scale=sqrt(3)
    )
    return lengthscale_prior, lengthscale_prior.mode.item()


class DefaultRBFKernel(RBFKernel):
    """A simple wrapper around RBF kernel that integrates the dim-scaled
    prior commonly used by default in BoTorch models, if mle=False.

    This supports easy configuration of the kernel with options like
    `active_dims`, while avoiding the need to explicitly specify the
    prior for the RBF kernel.
    """

    def __init__(
        self,
        ard_num_dims: int | None,
        active_dims: Sequence[int] | None = None,
        batch_shape: torch.Size | None = None,
        mle: bool = False,
    ) -> None:
        """Initialize Matern kernel with dimension-scaling prior or MLE.

        Args:
            ard_num_dims: The number of ARD dimensions. None signifies that an
                isotropic kernel should be used
            active_dims: The active input dimensions.
            batch_shape: The batch shape for the kernel.
            mle: A boolean indicating whether to use MLE (no priors) or a dimension
                scaling prior.
        """
        lengthscale_prior, initial_value = get_lengthscale_prior_and_initial_value(
            ard_num_dims=ard_num_dims, mle=mle
        )
        super().__init__(
            ard_num_dims=ard_num_dims,
            batch_shape=batch_shape,
            lengthscale_prior=lengthscale_prior,
            lengthscale_constraint=GreaterThan(
                2.5e-2, transform=None, initial_value=initial_value
            ),
            # pyre-ignore[6] GPyTorch type is unnecessarily restrictive.
            active_dims=active_dims,
        )


class DefaultMaternKernel(MaternKernel):
    """A simple wrapper arounda Matern kernel that integrates the dim-scaled
    prior commonly used by default in BoTorch models (if mle is false).

    This supports easy configuration of the kernel with options like
    `active_dims`, while avoiding the need to explicitly specify the
    prior for the RBF kernel.
    """

    def __init__(
        self,
        ard_num_dims: int | None,
        active_dims: Sequence[int] | None = None,
        batch_shape: torch.Size | None = None,
        nu: float = 2.5,
        mle: bool = False,
    ) -> None:
        lengthscale_prior, initial_value = get_lengthscale_prior_and_initial_value(
            ard_num_dims=ard_num_dims, mle=mle
        )
        super().__init__(
            ard_num_dims=ard_num_dims,
            batch_shape=batch_shape,
            lengthscale_prior=lengthscale_prior,
            lengthscale_constraint=GreaterThan(
                2.5e-2, transform=None, initial_value=initial_value
            ),
            active_dims=active_dims,
            nu=nu,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from ax.api.utils.instantiation.from_string import (
    optimization_config_from_string,
    parse_objective,
    parse_outcome_constraint,
    parse_parameter_constraint,
)
from ax.core.map_metric import MapMetric
from ax.core.objective import MultiObjective, Objective, ScalarizedObjective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import (
    ComparisonOp,
    ObjectiveThreshold,
    OutcomeConstraint,
    ScalarizedOutcomeConstraint,
)
from ax.core.parameter_constraint import ParameterConstraint
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase


class TestFromString(TestCase):
    def test_optimization_config_from_string(self) -> None:
        only_objective = optimization_config_from_string(objective_str="ne")
        self.assertEqual(
            only_objective,
            OptimizationConfig(
                objective=Objective(
                    metric=MapMetric(name="ne", lower_is_better=False), minimize=False
                ),
            ),
        )

        with_constraints = optimization_config_from_string(
            objective_str="ne", outcome_constraint_strs=["qps >= 0"]
        )
        self.assertEqual(
            with_constraints,
            OptimizationConfig(
                objective=Objective(
                    metric=MapMetric(name="ne", lower_is_better=False), minimize=False
                ),
                outcome_constraints=[
                    OutcomeConstraint(
                        metric=MapMetric(name="qps"),
                        op=ComparisonOp.GEQ,
                        bound=0.0,
                        relative=False,
                    )
                ],
            ),
        )

        with_constraints_and_objective_threshold = optimization_config_from_string(
            objective_str="-ne, qps",
            outcome_constraint_strs=["qps >= 1000", "flops <= 1000000"],
        )
        self.assertEqual(
            with_constraints_and_objective_threshold,
            MultiObjectiveOptimizationConfig(
                objective=MultiObjective(
                    objectives=[
                        Objective(
                            metric=MapMetric(name="ne", lower_is_better=True),
                            minimize=True,
                        ),
                        Objective(
                            metric=MapMetric(name="qps", lower_is_better=False),
                            minimize=False,
                        ),
                    ]
                ),
                outcome_constraints=[
                    OutcomeConstraint(
                        metric=MapMetric(name="flops"),
                        op=ComparisonOp.LEQ,
                        bound=1000000.0,
                        relative=False,
                    )
                ],
                objective_thresholds=[
                    ObjectiveThreshold(
                        metric=MapMetric(name="qps"),
                        op=ComparisonOp.GEQ,
                        bound=1000.0,
                        relative=False,
                    )
                ],
            ),
        )

    def test_parse_parameter_constraint(self) -> None:
        constraint = parse_parameter_constraint(constraint_str="x1 + x2 <= 1")
        self.assertEqual(
            constraint,
            ParameterConstraint(constraint_dict={"x1": 1, "x2": 1}, bound=1.0),
        )

        with_coefficients = parse_parameter_constraint(
            constraint_str="2 * x1 + 3 * x2 <= 1"
        )
        self.assertEqual(
            with_coefficients,
            ParameterConstraint(constraint_dict={"x1": 2, "x2": 3}, bound=1.0),
        )

        flipped_sign = parse_parameter_constraint(constraint_str="x1 + x2 >= 1")
        self.assertEqual(
            flipped_sign,
            ParameterConstraint(constraint_dict={"x1": -1, "x2": -1}, bound=-1.0),
        )

        weird = parse_parameter_constraint(constraint_str="x1 + x2 <= 1.5 * x3 + 2")
        self.assertEqual(
            weird,
            ParameterConstraint(
                constraint_dict={"x1": 1, "x2": 1, "x3": -1.5}, bound=2.0
            ),
        )

        with self.assertRaisesRegex(UserInputError, "Only linear"):
            parse_parameter_constraint(constraint_str="x1 * x2 <= 1")
        # test with sanitization
        constraint = parse_parameter_constraint(constraint_str="foo.bar + foo.baz <= 1")
        self.assertEqual(
            constraint,
            ParameterConstraint(
                constraint_dict={"foo.bar": 1, "foo.baz": 1}, bound=1.0
            ),
        )

    def test_parse_objective(self) -> None:
        single_objective = parse_objective(objective_str="ne")
        self.assertEqual(
            single_objective,
            Objective(
                metric=MapMetric(name="ne", lower_is_better=False), minimize=False
            ),
        )

        maximize_single_objective = parse_objective(objective_str="-qps")
        self.assertEqual(
            maximize_single_objective,
            Objective(
                metric=MapMetric(name="qps", lower_is_better=True), minimize=True
            ),
        )

        scalarized_objective = parse_objective(
            objective_str="0.5 * ne1 + 0.3 * ne2 + 0.2 * ne3"
        )
        self.assertEqual(
            scalarized_objective,
            ScalarizedObjective(
                metrics=[
                    MapMetric(name="ne1"),
                    MapMetric(name="ne2"),
                    MapMetric(name="ne3"),
                ],
                weights=[0.5, 0.3, 0.2],
                minimize=False,
            ),
        )

        multiobjective = parse_objective(objective_str="ne, -qps")
        self.assertEqual(
            multiobjective,
            MultiObjective(
                objectives=[
                    Objective(
                        metric=MapMetric(name="ne", lower_is_better=False),
                        minimize=False,
                    ),
                    Objective(
                        metric=MapMetric(name="qps", lower_is_better=True),
                        minimize=True,
                    ),
                ]
            ),
        )

        with self.assertRaisesRegex(UserInputError, "Only linear"):
            parse_objective(objective_str="ne * qps")

    def test_parse_outcome_constraint(self) -> None:
        constraint = parse_outcome_constraint(constraint_str="flops <= 1000000")
        self.assertEqual(
            constraint,
            OutcomeConstraint(
                metric=MapMetric(name="flops"),
                op=ComparisonOp.LEQ,
                bound=1000000.0,
                relative=False,
            ),
        )

        flipped_sign = parse_outcome_constraint(constraint_str="flops >= 1000000.0")
        self.assertEqual(
            flipped_sign,
            OutcomeConstraint(
                metric=MapMetric(name="flops"),
                op=ComparisonOp.GEQ,
                bound=1000000.0,
                relative=False,
            ),
        )

        relative = parse_outcome_constraint(constraint_str="flops <= 105 * baseline")
        self.assertEqual(
            relative,
            OutcomeConstraint(
                metric=MapMetric(name="flops"),
                op=ComparisonOp.LEQ,
                bound=105.0,
                relative=True,
            ),
        )

        scalarized = parse_outcome_constraint(
            constraint_str="0.5 * flops1 + 0.3 * flops2 <= 1000000"
        )
        self.assertEqual(
            scalarized,
            ScalarizedOutcomeConstraint(
                metrics=[MapMetric(name="flops1"), MapMetric(name="flops2")],
                weights=[0.5, 0.3],
                op=ComparisonOp.LEQ,
                bound=1000000.0,
                relative=False,
            ),
        )

        with self.assertRaisesRegex(UserInputError, "Expected an inequality"):
            parse_outcome_constraint(constraint_str="flops == 1000000")

        with self.assertRaisesRegex(UserInputError, "Only linear"):
            parse_outcome_constraint(constraint_str="flops * flops <= 1000000")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from typing import Mapping, Sequence

import pandas as pd
from ax.adapter.base import Adapter
from ax.adapter.cross_validation import cross_validate, CVResult
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardBase
from ax.analysis.plotly.color_constants import AX_BLUE

from ax.analysis.plotly.plotly_analysis import create_plotly_analysis_card

from ax.analysis.plotly.utils import get_scatter_point_color, Z_SCORE_95_CI

from ax.analysis.utils import extract_relevant_adapter
from ax.core.experiment import Experiment
from ax.generation_strategy.generation_strategy import GenerationStrategy
from plotly import graph_objects as go

CV_CARDGROUP_TITLE = "Cross Validation: Assessing model fit"

CV_CARDGROUP_SUBTITLE = (
    "Cross-validation plots display the model fit for each metric in the "
    "experiment. The model is trained on a subset of the data and then predicts the "
    "outcome for the remaining subset. The plots show the predicted outcome for the "
    "validation set on the y-axis against its actual value on the x-axis. Points "
    "that align closely with the dotted diagonal line indicate a strong model fit, "
    "signifying accurate predictions. Additionally, the plots include "
    "confidence intervals that provide insight into the noise in observations and "
    "the uncertainty in model predictions. <br><br>"
    "NOTE: A horizontal, flat line of predictions "
    "indicates that the model has not picked up on sufficient signal in the data, "
    "and instead is just predicting the mean."
)


class CrossValidationPlot(Analysis):
    """
    Plotly Scatter plot for cross validation for model predictions using the current
    model on the GenerationStrategy. This plot is useful for understanding how well
    the model is able to predict out-of-sample which in turn is indicative of its
    ability to suggest valuable candidates.

    Splits the model's training data into train/test folds and makes
    out-of-sample predictions on the test folds.

    A well fit model will have points clustered around the y=x line, and a model with
    poor fit may have points in a horizontal band in the center of the plot
    indicating a tendency to predict the observed mean of the specificed metric for
    all arms.

    The DataFrame computed will contain one row per arm and the following columns:
        - arm_name: The name of the arm
        - observed: The observed mean of the metric specified
        - observed_sem: The SEM of the observed mean of the metric specified
        - predicted: The predicted mean of the metric specified
        - predicted_sem: The SEM of the predicted mean of the metric specified
    """

    def __init__(
        self,
        metric_names: Sequence[str] | None = None,
        folds: int = -1,
        untransform: bool = False,
        trial_index: int | None = None,
        labels: Mapping[str, str] | None = None,
    ) -> None:
        """
        Args:
            metric_names: The name of the metric to plot. If not specified all metrics
                available on the underyling model will be used.
            folds: Number of subsamples to partition observations into. Use -1 for
                leave-one-out cross validation.
            untransform: Whether to untransform the model predictions before cross
                validating. Generators are trained on transformed data, and candidate
                generation is performed in the transformed space. Computing the model
                quality metric based on the cross-validation results in the
                untransformed space may not be representative of the model that
                is actually used for candidate generation in case of non-invertible
                transforms, e.g., Winsorize or LogY. While the model in the
                transformed space may not be representative of the original data in
                regions where outliers have been removed, we have found it to better
                reflect the how good the model used for candidate generation actually
                is.
            trial_index: Optional trial index that the model from generation_strategy
                was used to generate. Useful card attribute to filter to only specific
                trial.
            labels: Optional dictionary of labels for the plot. Useful for when metric
                names are too long or otherwise challenging to read.
        """

        self.metric_names = metric_names
        self.folds = folds
        self.untransform = untransform
        self.trial_index = trial_index
        self.labels: dict[str, str] = {**labels} if labels is not None else {}

    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardBase:
        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        cards = []
        cv_results = cross_validate(
            model=relevant_adapter, folds=self.folds, untransform=self.untransform
        )
        for metric_name in self.metric_names or relevant_adapter.metric_names:
            df = _prepare_data(metric_name=metric_name, cv_results=cv_results)

            fig = _prepare_plot(df=df)

            k_folds_substring = (
                f"{self.folds}-fold" if self.folds > 0 else "leave-one-out"
            )

            # If a human readable metric name is provided, use it in the title
            metric_title = self.labels.get(metric_name, metric_name)

            # Define the cross-validation description based on the number of folds
            cv_description = (
                (
                    f"the data is split into {self.folds} subsets and the model is "
                    f"trained on {self.folds - 1} subsets while the remaining subset "
                    "is used for validation"
                )
                if self.folds > 0
                else (
                    "the model is trained on all data except one sample, which is "
                    "used for validation"
                )
            )

            card = create_plotly_analysis_card(
                name=self.__class__.__name__,
                title=f"Cross Validation for {metric_title}",
                subtitle=(
                    "The cross-validation plot displays the model fit for each "
                    f"metric in the experiment. It employs a {k_folds_substring} "
                    f"approach, where {cv_description}. The plot shows the "
                    "predicted outcome for the validation set on the y-axis against "
                    "its actual value on the x-axis. Points that align closely with "
                    "the dotted diagonal line indicate a strong model fit, signifying "
                    "accurate predictions. Additionally, the plot includes 95% "
                    "confidence intervals that provide insight into the noise in "
                    "observations and the uncertainty in model predictions. A "
                    "horizontal, flat line of predictions indicates that the model "
                    "has not picked up on sufficient signal in the data, and instead "
                    "is just predicting the mean."
                ),
                df=df,
                fig=fig,
            )

            cards.append(card)

        return self._create_analysis_card_group_or_card(
            title=CV_CARDGROUP_TITLE,
            subtitle=CV_CARDGROUP_SUBTITLE,
            children=cards,
        )


def compute_cross_validation_adhoc(
    metric_names: Sequence[str] | None = None,
    folds: int = -1,
    untransform: bool = True,
    labels: Mapping[str, str] | None = None,
    experiment: Experiment | None = None,
    generation_strategy: GenerationStrategy | None = None,
    adapter: Adapter | None = None,
) -> AnalysisCardBase:
    """
    Helper method to expose adhoc cross validation plotting. Only for advanced users in
    a notebook setting.

    Args:
        metric_names: The name of the metric to plot. If not specified all metrics
            available on the underyling model will be used.
        folds: Number of subsamples to partition observations into. Use -1 for
            leave-one-out cross validation.
        untransform: Whether to untransform the model predictions before cross
            validating. Generators are trained on transformed data, and candidate
            generation is performed in the transformed space. Computing the model
            quality metric based on the cross-validation results in the
            untransformed space may not be representative of the model that
            is actually used for candidate generation in case of non-invertible
            transforms, e.g., Winsorize or LogY. While the model in the
            transformed space may not be representative of the original data in
            regions where outliers have been removed, we have found it to better
            reflect the how good the model used for candidate generation actually
            is.
        labels: Optional dictionary of labels for the plot. Useful for when metric
            names are too long or otherwise challenging to read.
        experiment: Optional. The experiment to extract data from.
        generation_strategy: Optional. The generation strategy to extract the adapter
            from.
        adapter: Optional. The adapter to cross validate. If provided, this adapter
            will be used instead of the current adapter on the ``GenerationStrategy``
    """
    relevant_adapter = extract_relevant_adapter(
        experiment=experiment,
        generation_strategy=generation_strategy,
        adapter=adapter,
    )

    analysis = CrossValidationPlot(
        metric_names=metric_names,
        folds=folds,
        untransform=untransform,
        labels=labels,
    )

    return analysis.compute(
        experiment=experiment,
        adapter=relevant_adapter,
    )


def _prepare_data(metric_name: str, cv_results: list[CVResult]) -> pd.DataFrame:
    records = []
    for observed, predicted in cv_results:
        # Find the index of the metric in observed and predicted
        observed_i = next(
            (
                i
                for i, name in enumerate(observed.data.metric_names)
                if name == metric_name
            ),
            None,
        )
        predicted_i = next(
            (i for i, name in enumerate(predicted.metric_names) if name == metric_name),
            None,
        )
        # Check if both indices are found
        if observed_i is not None and predicted_i is not None:
            record = {
                "arm_name": observed.arm_name,
                "observed": observed.data.means[observed_i],
                "predicted": predicted.means[predicted_i],
                # Compute the 95% confidence intervals for plotting purposes
                "observed_95_ci": observed.data.covariance[observed_i][observed_i]
                ** 0.5
                * Z_SCORE_95_CI,
                "predicted_95_ci": predicted.covariance[predicted_i][predicted_i] ** 0.5
                * Z_SCORE_95_CI,
            }
            records.append(record)
    return pd.DataFrame.from_records(records)


def _prepare_plot(
    df: pd.DataFrame,
) -> go.Figure:
    # Create a scatter plot using Plotly Graph Objects for more control
    fig = go.Figure()
    TRANSPARENT_AX_BLUE: str = get_scatter_point_color(
        hex_color=AX_BLUE,
        ci_transparency=True,
    )
    FILLED_AX_BLUE: str = get_scatter_point_color(
        hex_color=AX_BLUE,
        ci_transparency=False,
    )
    fig.add_trace(
        go.Scatter(
            x=df["observed"],
            y=df["predicted"],
            mode="markers",
            marker={
                "color": FILLED_AX_BLUE,
            },
            error_x={
                "type": "data",
                "array": df["observed_95_ci"],
                "visible": True,
                "color": TRANSPARENT_AX_BLUE,
            },
            error_y={
                "type": "data",
                "array": df["predicted_95_ci"],
                "visible": True,
                "color": TRANSPARENT_AX_BLUE,
            },
            text=df["arm_name"],
            hovertemplate=(
                "<b>Arm Name: %{text}</b><br>"
                + "Predicted: %{y}<br>"
                + "Observed: %{x}<br>"
                + "<extra></extra>"  # Removes the trace name from the hover
            ),
            hoverlabel={
                "bgcolor": TRANSPARENT_AX_BLUE,
                "font": {"color": "black"},
            },
        )
    )

    # Add a gray dashed line at y=x starting and ending just outside of the region of
    # interest for reference. A well fit model should have points clustered around
    # this line.
    lower_bound = (
        min(
            (df["observed"] - df["observed_95_ci"].fillna(0)).min(),
            (df["predicted"] - df["predicted_95_ci"].fillna(0)).min(),
        )
        * 0.999  # tight autozoom
    )
    upper_bound = (
        max(
            (df["observed"] + df["observed_95_ci"].fillna(0)).max(),
            (df["predicted"] + df["predicted_95_ci"].fillna(0)).max(),
        )
        * 1.001  # tight autozoom
    )
    fig.add_shape(
        type="line",
        x0=lower_bound,
        y0=lower_bound,
        x1=upper_bound,
        y1=upper_bound,
        line={"color": "gray", "dash": "dot"},
    )

    # Update axes with tight autozoom that remains square
    fig.update_xaxes(
        range=[lower_bound, upper_bound], constrain="domain", title="Actual Outcome"
    )
    fig.update_yaxes(
        range=[lower_bound, upper_bound],
        scaleanchor="x",
        scaleratio=1,
        title="Predicted Outcome",
    )
    return fig



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import os

from ax.benchmark.benchmark_metric import BenchmarkMetric
from ax.benchmark.benchmark_problem import BenchmarkProblem
from ax.benchmark.problems.surrogate.hss.base import HierarchicalSearchSpaceSurrogate
from ax.core.objective import Objective
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.search_space import HierarchicalSearchSpace
from xgboost import XGBRegressor


# The optimal value is obtained by a dense grid search, whereas the baseline value is
# obtained by `compute_baseline_value_from_sobol`
MNIST_SURROGATE_OPTIMAL_VALUE = 1.0790634
MNIST_SURROGATE_BASELINE_VALUE = 0.9837410950660705


def get_mnist_surrogate_search_space() -> HierarchicalSearchSpace:
    search_space = HierarchicalSearchSpace(
        parameters=[
            FixedParameter(
                name="root",
                value=True,
                parameter_type=ParameterType.BOOL,
                dependents={
                    True: ["lr", "use_dropout", "use_weight_decay"],
                },
            ),
            RangeParameter(
                name="lr",
                parameter_type=ParameterType.FLOAT,
                lower=1e-5,
                upper=1,
                log_scale=True,
            ),
            ChoiceParameter(
                "use_dropout",
                parameter_type=ParameterType.BOOL,
                values=[False, True],
                is_ordered=True,
                sort_values=True,
                dependents={
                    False: [],
                    True: ["dropout"],
                },
            ),
            RangeParameter(
                name="dropout",
                parameter_type=ParameterType.FLOAT,
                lower=0.1,
                upper=0.5,
            ),
            ChoiceParameter(
                "use_weight_decay",
                parameter_type=ParameterType.BOOL,
                values=[False, True],
                is_ordered=True,
                sort_values=True,
                dependents={
                    False: [],
                    True: ["weight_decay"],
                },
            ),
            RangeParameter(
                name="weight_decay",
                parameter_type=ParameterType.FLOAT,
                lower=1e-5,
                upper=1,
                log_scale=True,
            ),
        ],
    )

    return search_space


def get_mnist_surrogate_arguments() -> (
    tuple[list[list[str]], list[str], list[dict[str, bool]], list[XGBRegressor]]
):
    lst_active_param_names: list[list[str]] = [
        ["lr"],
        ["lr", "weight_decay"],
        ["lr", "dropout"],
        ["lr", "dropout", "weight_decay"],
    ]

    flag_param_names: list[str] = ["use_dropout", "use_weight_decay"]

    lst_flag_config: list[dict[str, bool]] = [
        {"use_dropout": False, "use_weight_decay": False},
        {"use_dropout": False, "use_weight_decay": True},
        {"use_dropout": True, "use_weight_decay": False},
        {"use_dropout": True, "use_weight_decay": True},
    ]

    lst_xgb_models: list[XGBRegressor] = [XGBRegressor() for i in range(4)]

    for i, xgb_model in enumerate(lst_xgb_models):
        xgb_model.load_model(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "checkpoints",
                f"mnist_xgb_model_{i:d}.json",
            )
        )

    return lst_active_param_names, flag_param_names, lst_flag_config, lst_xgb_models


def get_mnist_surrogate_benchmark(
    num_trials: int = 50,
) -> BenchmarkProblem:
    """Creates a BenchmarkProblem for tuning hyperparameters of a MNIST convolutional
    neural network.

    The hyperparameters are the learning rate, dropout rate, and weight decay. The
    dropout rate and the weight decay coefficient are optional hyperparameters, and can
    be turned off. This benchmark problem uses a surrogate model, i.e., gradient
    boosted trees, to predict the validation accuracy, which makes the oracle
    computationally cheap to evaluate.
    """
    search_space = get_mnist_surrogate_search_space()

    lst_active_param_names, flag_param_names, lst_flag_config, lst_xgb_models = (
        get_mnist_surrogate_arguments()
    )

    test_function = HierarchicalSearchSpaceSurrogate(
        outcome_names=["MNIST Test Accuracy"],
        lst_active_param_names=lst_active_param_names,
        flag_param_names=flag_param_names,
        lst_flag_config=lst_flag_config,
        lst_xgb_models=lst_xgb_models,
    )

    optimization_config = OptimizationConfig(
        objective=Objective(
            metric=BenchmarkMetric(
                name="MNIST Test Accuracy",
                lower_is_better=False,
                observe_noise_sd=False,
            ),
            minimize=False,
        )
    )

    return BenchmarkProblem(
        name="MNIST Test Accuracy",
        search_space=search_space,
        optimization_config=optimization_config,
        num_trials=num_trials,
        optimal_value=MNIST_SURROGATE_OPTIMAL_VALUE,
        test_function=test_function,
        baseline_value=MNIST_SURROGATE_BASELINE_VALUE,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import copy
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from typing import Any

from ax.benchmark.benchmark_problem import BenchmarkProblem
from ax.benchmark.problems.runtime_funcs import int_from_params
from ax.benchmark.problems.surrogate.lcbench.early_stopping import (
    get_lcbench_early_stopping_benchmark_problem,
)
from ax.benchmark.problems.surrogate.lcbench.transfer_learning import (
    get_lcbench_benchmark_problem,
)
from ax.benchmark.problems.synthetic.bandit import get_bandit_problem
from ax.benchmark.problems.synthetic.discretized.mixed_integer import (
    get_discrete_ackley,
    get_discrete_hartmann,
    get_discrete_rosenbrock,
)
from ax.benchmark.problems.synthetic.from_botorch import create_problem_from_botorch
from ax.benchmark.problems.synthetic.hss.jenatton import get_jenatton_benchmark_problem
from botorch.test_functions import synthetic
from botorch.test_functions.multi_objective import BraninCurrin


@dataclass
class BenchmarkProblemRegistryEntry:
    factory_fn: Callable[..., BenchmarkProblem]
    factory_kwargs: dict[str, Any]


# Baseline values were obtained with `compute_baseline_value_from_sobol`
BENCHMARK_PROBLEM_REGISTRY: dict[str, BenchmarkProblemRegistryEntry] = {
    "ackley4": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Ackley,
            "test_problem_kwargs": {"dim": 4},
            "num_trials": 40,
            "observe_noise_sd": False,
        },
    ),
    "ackley4_async_noisy": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Ackley,
            "test_problem_kwargs": {"dim": 4},
            "num_trials": 40,
            "noise_std": 1.0,
            "observe_noise_sd": False,
            "step_runtime_function": int_from_params,
            "name": "ackley4_async_noisy",
        },
    ),
    "Bandit": BenchmarkProblemRegistryEntry(
        factory_fn=get_bandit_problem, factory_kwargs={}
    ),
    "branin": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Branin,
            "test_problem_kwargs": {},
            "num_trials": 30,
            "observe_noise_sd": False,
        },
    ),
    "branin_currin": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": BraninCurrin,
            "test_problem_kwargs": {},
            "num_trials": 30,
            "observe_noise_sd": False,
        },
    ),
    "branin_currin30": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": BraninCurrin,
            "test_problem_kwargs": {},
            "num_trials": 30,
            "observe_noise_sd": False,
            "baseline_value": 3.0187520516793587,
            "n_dummy_dimensions": 28,
        },
    ),
    "griewank4": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Griewank,
            "test_problem_kwargs": {"dim": 4},
            "num_trials": 40,
            "observe_noise_sd": False,
        },
    ),
    "hartmann3": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Hartmann,
            "test_problem_kwargs": {"dim": 3},
            "num_trials": 30,
            "observe_noise_sd": False,
        },
    ),
    "hartmann6": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Hartmann,
            "test_problem_kwargs": {"dim": 6},
            "num_trials": 35,
            "observe_noise_sd": False,
        },
    ),
    "hartmann30": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Hartmann,
            "test_problem_kwargs": {"dim": 6},
            "num_trials": 25,
            "observe_noise_sd": False,
            "n_dummy_dimensions": 24,
        },
    ),
    "jenatton": BenchmarkProblemRegistryEntry(
        factory_fn=get_jenatton_benchmark_problem,
        factory_kwargs={"num_trials": 50, "observe_noise_sd": False},
    ),
    "LCBench:v1 Fashion-MNIST": BenchmarkProblemRegistryEntry(
        get_lcbench_benchmark_problem, factory_kwargs={"dataset_name": "Fashion-MNIST"}
    ),
    "levy4": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Levy,
            "test_problem_kwargs": {"dim": 4},
            "num_trials": 40,
            "observe_noise_sd": False,
        },
    ),
    "powell4": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Powell,
            "test_problem_kwargs": {"dim": 4},
            "num_trials": 40,
            "observe_noise_sd": False,
        },
    ),
    "rosenbrock4": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Rosenbrock,
            "test_problem_kwargs": {"dim": 4},
            "num_trials": 40,
            "observe_noise_sd": False,
        },
    ),
    "six_hump_camel": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.SixHumpCamel,
            "test_problem_kwargs": {},
            "num_trials": 30,
            "observe_noise_sd": False,
        },
    ),
    "three_hump_camel": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.ThreeHumpCamel,
            "test_problem_kwargs": {},
            "num_trials": 30,
            "observe_noise_sd": False,
        },
    ),
    # Problems where we observe the noise level
    "branin_observed_noise": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Branin,
            "test_problem_kwargs": {},
            "num_trials": 20,
            "observe_noise_sd": True,
        },
    ),
    "branin_currin_observed_noise": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": BraninCurrin,
            "test_problem_kwargs": {},
            "num_trials": 30,
            "observe_noise_sd": True,
        },
    ),
    "branin_currin30_observed_noise": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": BraninCurrin,
            "test_problem_kwargs": {},
            "num_trials": 30,
            "observe_noise_sd": True,
            "n_dummy_dimensions": 28,
        },
    ),
    "hartmann6_observed_noise": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Hartmann,
            "test_problem_kwargs": {"dim": 6},
            "num_trials": 50,
            "observe_noise_sd": True,
        },
    ),
    "hartmann30_observed_noise": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.Hartmann,
            "test_problem_kwargs": {"dim": 6},
            "observe_noise_sd": True,
            "num_trials": 25,
            "n_dummy_dimensions": 24,
        },
    ),
    "jenatton_observed_noise": BenchmarkProblemRegistryEntry(
        factory_fn=get_jenatton_benchmark_problem,
        factory_kwargs={"num_trials": 25, "observe_noise_sd": True},
    ),
    "constrained_gramacy_observed_noise": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.ConstrainedGramacy,
            "test_problem_kwargs": {},
            "num_trials": 50,
            "observe_noise_sd": True,
        },
    ),
    "Discrete Hartmann": BenchmarkProblemRegistryEntry(
        factory_fn=get_discrete_hartmann,
        factory_kwargs={},
    ),
    "Discrete Ackley": BenchmarkProblemRegistryEntry(
        factory_fn=get_discrete_ackley, factory_kwargs={}
    ),
    "Discrete Rosenbrock": BenchmarkProblemRegistryEntry(
        factory_fn=get_discrete_rosenbrock, factory_kwargs={}
    ),
    "PressureVessel": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.PressureVessel,
            "test_problem_kwargs": {},
            "num_trials": 50,
        },
    ),
    "TensionCompressionString": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.TensionCompressionString,
            "test_problem_kwargs": {},
            "num_trials": 50,
        },
    ),
    "WeldedBeamSO": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.WeldedBeamSO,
            "test_problem_kwargs": {},
            "num_trials": 50,
        },
    ),
    "SpeedReducer": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.SpeedReducer,
            "test_problem_kwargs": {},
            "num_trials": 50,
        },
    ),
    "KeaneBumpFunction2": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.KeaneBumpFunction,
            "test_problem_kwargs": {"dim": 2},
            "num_trials": 30,
        },
    ),
    "KeaneBumpFunction10": BenchmarkProblemRegistryEntry(
        factory_fn=create_problem_from_botorch,
        factory_kwargs={
            "test_problem_class": synthetic.KeaneBumpFunction,
            "test_problem_kwargs": {"dim": 10},
            "num_trials": 50,
        },
    ),
    **{
        f"LCBench:ES {dataset_name}": BenchmarkProblemRegistryEntry(
            get_lcbench_early_stopping_benchmark_problem,
            factory_kwargs={"dataset_name": dataset_name},
        )
        for dataset_name in (
            "airlines",
            "albert",
            "covertype",
            "christine",
            "Fashion-MNIST",
        )
    },
}


def get_benchmark_problem(
    problem_key: str,
    registry: Mapping[str, BenchmarkProblemRegistryEntry] | None = None,
    **additional_kwargs: Any,
) -> BenchmarkProblem:
    """
    Generate a benchmark problem from a key, registry, and additional arguments.

    Args:
        problem_key: The key by which a `BenchmarkProblemRegistryEntry` is
            looked up in the registry; a problem will then be generated from
            that entry and `additional_kwargs`. Note that this is not
            necessarily the same as the `name` attribute of the problem, and
            that one `problem_key` can generate several different
            `BenchmarkProblem`s by passing `additional_kwargs`. However, it is a
            good practice to maintain a 1:1 mapping between `problem_key` and
            the name.
        registry: If not provided, uses `BENCHMARK_PROBLEM_REGISTRY` to use
            problems defined within Ax.
        additional_kwargs: Additional kwargs to pass to the factory function of
            the `BenchmarkProblemRegistryEntry`.
    """
    registry = BENCHMARK_PROBLEM_REGISTRY if registry is None else registry
    entry = registry[problem_key]
    kwargs = copy.copy(entry.factory_kwargs)
    kwargs.update(additional_kwargs)
    return entry.factory_fn(**kwargs)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import dataclasses
from collections import OrderedDict
from collections.abc import Sequence
from logging import Logger
from typing import Any

import numpy.typing as npt
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.core.types import TCandidateMetadata, TGenMetadata
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.exceptions.model import ModelError
from ax.generators.torch.botorch import (
    get_feature_importances_from_botorch_model,
    get_rounding_func,
)
from ax.generators.torch.botorch_modular.acquisition import Acquisition
from ax.generators.torch.botorch_modular.multi_acquisition import MultiAcquisition
from ax.generators.torch.botorch_modular.surrogate import Surrogate, SurrogateSpec
from ax.generators.torch.botorch_modular.utils import (
    choose_botorch_acqf_class,
    construct_acquisition_and_optimizer_options,
    ModelConfig,
)
from ax.generators.torch.utils import _to_inequality_constraints
from ax.generators.torch_base import TorchGenerator, TorchGenResults, TorchOptConfig
from ax.utils.common.base import Base
from ax.utils.common.constants import Keys
from ax.utils.common.docutils import copy_doc
from ax.utils.common.logger import get_logger
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.models.deterministic import FixedSingleSampleModel
from botorch.settings import validate_input_scaling
from botorch.utils.datasets import SupervisedDataset
from pyre_extensions import assert_is_instance, none_throws
from torch import Tensor

logger: Logger = get_logger(__name__)


class BoTorchGenerator(TorchGenerator, Base):
    """**All classes in 'botorch_modular' directory are under
    construction, incomplete, and should be treated as alpha
    versions only.**

    Modular ``Model`` class for combining BoTorch subcomponents
    in Ax. Specified via ``Surrogate`` and ``Acquisition``, which wrap
    BoTorch ``Model`` and ``AcquisitionFunction``, respectively, for
    convenient use in Ax.

    Args:
        acquisition_class: Type of ``Acquisition`` to be used in
            this model, auto-selected based on experiment and data
            if not specified.
        acquisition_options: Optional dict of kwargs, passed to
            the the Ax ``Acquisition`` class.
        botorch_acqf_class: Type of ``AcquisitionFunction`` to be
            used in this model, auto-selected based on experiment
            and data if not specified.
        botorch_acqf_options: Optional dict of kwargs, passed to the botorch
            ``AcquisitionFunction``.
        botorch_acqf_classes_with_options: List of tuples of
            ``AcquisitionFunction`` classes and dicts of kwargs, passed to
            the botorch ``AcquisitionFunction``. This is used to specify
            multiple acquisition functions to be used with MultiAcquisition.
        surrogate_spec: An optional ``SurrogateSpec`` object specifying how to
            construct the ``Surrogate`` and the underlying BoTorch ``Model``.
        surrogate: In lieu of ``SurrogateSpec``, an instance of ``Surrogate`` may
            be provided. In most cases, ``surrogate_spec`` should be used instead.
        refit_on_cv: Whether to reoptimize model parameters during call to
            ``BoTorchGenerator.cross_validate``.
        warm_start_refit: Whether to load parameters from either the provided
            state dict or the state dict of the current BoTorch ``Model`` during
            refitting. If False, model parameters will be reoptimized from
            scratch on refit. NOTE: This setting is ignored during
            ``cross_validate`` if ``refit_on_cv`` is False. This is also used in
            Surrogate.model_selection.
        use_p_feasible: Whether we consider dispatching to
            ``qLogProbabilityOfFeasibility`` in ``choose_botorch_acqf_class``.
    """

    acquisition_class: type[Acquisition]
    acquisition_options: dict[str, Any]

    surrogate_spec: SurrogateSpec | None
    _surrogate: Surrogate | None

    _user_specified_botorch_acqf_class: type[AcquisitionFunction] | None
    _botorch_acqf_class: type[AcquisitionFunction] | None
    _botorch_acqf_options: dict[str, Any]
    _supports_robust_optimization: bool = True
    _acquisition: Acquisition | None = None

    def __init__(
        self,
        surrogate_spec: SurrogateSpec | None = None,
        surrogate: Surrogate | None = None,
        acquisition_class: type[Acquisition] | None = None,
        acquisition_options: dict[str, Any] | None = None,
        botorch_acqf_class: type[AcquisitionFunction] | None = None,
        botorch_acqf_options: dict[str, Any] | None = None,
        botorch_acqf_classes_with_options: list[
            tuple[type[AcquisitionFunction], dict[str, Any]]
        ]
        | None = None,
        refit_on_cv: bool = False,
        warm_start_refit: bool = True,
        use_p_feasible: bool = True,
    ) -> None:
        # Check that only one surrogate related option is provided.
        if surrogate_spec is not None and surrogate is not None:
            raise UserInputError(
                "Only one of `surrogate_spec` or `surrogate` "
                "can be specified. Please use `surrogate_spec`."
            )
        self.surrogate_spec = surrogate_spec
        self._surrogate = surrogate

        if botorch_acqf_class is not None:
            if botorch_acqf_classes_with_options is not None:
                raise UserInputError(
                    "Only one of `botorch_acqf_class` or "
                    "`botorch_acqf_classes_with_options` can be specified."
                )

        if (
            botorch_acqf_classes_with_options is not None
            and len(botorch_acqf_classes_with_options) >= 2
        ):
            if (
                acquisition_class is not None
                and acquisition_class is not MultiAcquisition
            ):
                raise UserInputError(
                    "Multiple classes in `botorch_acqf_classes_with_options`"
                    "must be used with MultiAcquisition."
                )
            acquisition_class = MultiAcquisition

        self.acquisition_class = acquisition_class or Acquisition

        self.acquisition_options = acquisition_options or {}
        self._user_specified_botorch_acqf_class = botorch_acqf_class
        self._botorch_acqf_class = botorch_acqf_class
        self._botorch_acqf_classes_with_options = botorch_acqf_classes_with_options
        self._botorch_acqf_options = botorch_acqf_options or {}

        self.refit_on_cv = refit_on_cv
        self.warm_start_refit = warm_start_refit
        self.use_p_feasible = use_p_feasible

    @property
    def surrogate(self) -> Surrogate:
        """Returns the ``Surrogate``, if it has been constructed."""
        if self._surrogate is None:
            raise ModelError("Surrogate has not yet been constructed.")
        return self._surrogate

    @property
    def Xs(self) -> list[Tensor]:
        """A list of tensors, each of shape ``batch_shape x n_i x d``,
        where `n_i` is the number of training inputs for the i-th model.

        NOTE: This is an accessor for ``self.surrogate.Xs``
        and returns it unchanged.
        """
        return self.surrogate.Xs

    @property
    def botorch_acqf_class(self) -> type[AcquisitionFunction]:
        """BoTorch ``AcquisitionFunction`` class, associated with this model.
        Raises an error if one is not yet set.
        """
        if not self._botorch_acqf_class:
            raise ValueError("BoTorch `AcquisitionFunction` has not yet been set.")
        return self._botorch_acqf_class

    def fit(
        self,
        datasets: Sequence[SupervisedDataset],
        search_space_digest: SearchSpaceDigest,
        candidate_metadata: list[list[TCandidateMetadata]] | None = None,
        state_dict: OrderedDict[str, Tensor] | None = None,
        refit: bool = True,
        **additional_model_inputs: Any,
    ) -> None:
        """Fit model to m outcomes.

        Args:
            datasets: A list of ``SupervisedDataset`` containers, each
                corresponding to the data of one or more outcomes.
            search_space_digest: A ``SearchSpaceDigest`` object containing
                metadata on the features in the datasets.
            candidate_metadata: Model-produced metadata for candidates, in
                the order corresponding to the Xs.
            state_dict: An optional model statedict for the underlying ``Surrogate``.
                Primarily used in ``BoTorchGenerator.cross_validate``.
            refit: Whether to re-optimize model parameters.
            additional_model_inputs: Additional kwargs to pass to the
                model input constructor in ``Surrogate.fit``.
        """
        # If a surrogate has not been constructed, construct it.
        if self._surrogate is None:
            surrogate_spec = (
                SurrogateSpec(model_configs=[ModelConfig(name="default")])
                if self.surrogate_spec is None
                else self.surrogate_spec
            )
            self._surrogate = Surrogate(
                surrogate_spec=surrogate_spec,
                refit_on_cv=self.refit_on_cv,
                warm_start_refit=self.warm_start_refit,
            )

        # Fit the surrogate.
        for config in self.surrogate.surrogate_spec.model_configs:
            config.model_options.update(additional_model_inputs)
        for (
            config_list
        ) in self.surrogate.surrogate_spec.metric_to_model_configs.values():
            for config in config_list:
                config.model_options.update(additional_model_inputs)
        self.surrogate.fit(
            datasets=datasets,
            search_space_digest=search_space_digest,
            candidate_metadata=candidate_metadata,
            state_dict=state_dict,
            refit=refit,
            repeat_model_selection_if_dataset_changed=True,
        )

    def predict(
        self, X: Tensor, use_posterior_predictive: bool = False
    ) -> tuple[Tensor, Tensor]:
        """Predicts, potentially from multiple surrogates.

        Args:
            X: (n x d) Tensor of input locations.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).

        Returns: Tuple of tensors: (n x m) mean, (n x m x m) covariance.
        """
        return self.surrogate.predict(
            X=X, use_posterior_predictive=use_posterior_predictive
        )

    @copy_doc(TorchGenerator.gen)
    def gen(
        self,
        n: int,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
    ) -> TorchGenResults:
        (
            acq_options,
            botorch_acqf_options,
            opt_options,
            botorch_acqf_classes_with_options,
        ) = construct_acquisition_and_optimizer_options(
            acqf_options=self.acquisition_options,
            botorch_acqf_options=self._botorch_acqf_options,
            model_gen_options=torch_opt_config.model_gen_options,
            botorch_acqf_classes_with_options=self._botorch_acqf_classes_with_options,
        )
        self._acquisition = self._instantiate_acquisition(
            search_space_digest=search_space_digest,
            torch_opt_config=torch_opt_config,
            botorch_acqf_options=botorch_acqf_options,
            acq_options=acq_options,
            botorch_acqf_classes_with_options=botorch_acqf_classes_with_options,
            n=n,
        )
        acqf = none_throws(self._acquisition)

        botorch_rounding_func = get_rounding_func(torch_opt_config.rounding_func)
        candidates, expected_acquisition_value, weights = acqf.optimize(
            n=n,
            search_space_digest=search_space_digest,
            inequality_constraints=_to_inequality_constraints(
                linear_constraints=torch_opt_config.linear_constraints
            ),
            fixed_features=torch_opt_config.fixed_features,
            rounding_func=botorch_rounding_func,
            optimizer_options=assert_is_instance(
                opt_options,
                dict,
            ),
        )
        gen_metadata = self._get_gen_metadata_from_acqf(
            acqf=acqf,
            torch_opt_config=torch_opt_config,
            expected_acquisition_value=expected_acquisition_value,
        )
        return TorchGenResults(
            points=candidates.detach().cpu(),
            weights=weights,
            gen_metadata=gen_metadata,
        )

    def _get_gen_metadata_from_acqf(
        self,
        acqf: Acquisition,
        torch_opt_config: TorchOptConfig,
        expected_acquisition_value: Tensor,
    ) -> TGenMetadata:
        gen_metadata: TGenMetadata = {
            Keys.EXPECTED_ACQF_VAL: expected_acquisition_value.tolist()
        }
        if torch_opt_config.objective_weights.nonzero().numel() > 1:
            gen_metadata["objective_thresholds"] = acqf.objective_thresholds
            gen_metadata["objective_weights"] = acqf.objective_weights

        if hasattr(acqf.acqf, "outcome_model"):
            outcome_model = acqf.acqf.outcome_model
            if isinstance(
                outcome_model,
                FixedSingleSampleModel,
            ):
                gen_metadata["outcome_model_fixed_draw_weights"] = outcome_model.w
        gen_metadata["models_used"] = acqf.models_used
        return gen_metadata

    @copy_doc(TorchGenerator.best_point)
    def best_point(
        self,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
    ) -> Tensor | None:
        try:
            return self.surrogate.best_in_sample_point(
                search_space_digest=search_space_digest,
                torch_opt_config=torch_opt_config,
            )[0]
        except ValueError:
            return None

    @copy_doc(TorchGenerator.evaluate_acquisition_function)
    def evaluate_acquisition_function(
        self,
        X: Tensor,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
        acq_options: dict[str, Any] | None = None,
        botorch_acqf_options: dict[str, Any] | None = None,
    ) -> Tensor:
        acqf = self._instantiate_acquisition(
            search_space_digest=search_space_digest,
            torch_opt_config=torch_opt_config,
            botorch_acqf_options=botorch_acqf_options or self._botorch_acqf_options,
            acq_options=acq_options,
        )
        return acqf.evaluate(X=X)

    @copy_doc(TorchGenerator.cross_validate)
    def cross_validate(
        self,
        datasets: Sequence[SupervisedDataset],
        X_test: Tensor,
        search_space_digest: SearchSpaceDigest,
        use_posterior_predictive: bool = False,
    ) -> tuple[Tensor, Tensor]:
        current_surrogate = self.surrogate
        # If we should be refitting but not warm-starting the refit, set
        # `state_dict` to None to avoid loading it.
        state_dict = (
            None
            if self.refit_on_cv and not self.warm_start_refit
            else current_surrogate.model.state_dict()
        )

        # Temporarily set `_surrogate` to cloned surrogate to set
        # the training data on cloned surrogate to train set and
        # use it to predict the test point.
        self._surrogate = current_surrogate.clone_reset()

        # Remove the `robust_digest` since we do not want to use perturbations here.
        search_space_digest = dataclasses.replace(
            search_space_digest,
            robust_digest=None,
        )

        try:
            # Since each CV fold removes points from the training data, the
            # remaining observations will not pass the input scaling checks.
            # To avoid confusing users with warnings, we disable these checks.
            with validate_input_scaling(False):
                self.surrogate.fit(
                    datasets=datasets,
                    search_space_digest=search_space_digest,
                    state_dict=state_dict,
                    refit=self.refit_on_cv,
                    repeat_model_selection_if_dataset_changed=False,
                )
            X_test_prediction = self.predict(
                X=X_test,
                use_posterior_predictive=use_posterior_predictive,
            )
        finally:
            # Reset the surrogates back to this model's surrogate, make
            # sure the cloned surrogate doesn't stay around if fit or
            # predict fail.
            self._surrogate = current_surrogate
        return X_test_prediction

    @property
    def dtype(self) -> torch.dtype:
        """Torch data type of the tensors in the training data used in the model,
        of which this ``Acquisition`` is a subcomponent.
        """
        return self.surrogate.dtype

    @property
    def device(self) -> torch.device:
        """Torch device type of the tensors in the training data used in the model,
        of which this ``Acquisition`` is a subcomponent.
        """
        return self.surrogate.device

    def _instantiate_acquisition(
        self,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
        botorch_acqf_options: dict[str, Any],
        acq_options: dict[str, Any] | None = None,
        botorch_acqf_classes_with_options: list[
            tuple[type[AcquisitionFunction], dict[str, Any]]
        ]
        | None = None,
        n: int | None = None,
    ) -> Acquisition:
        """Set a BoTorch acquisition function class for this model if needed and
        instantiate it.

        Returns:
            A BoTorch ``AcquisitionFunction`` instance.
        """
        if (
            self._user_specified_botorch_acqf_class is None
            and self._botorch_acqf_classes_with_options is None
        ):
            if torch_opt_config.risk_measure is not None:
                raise UnsupportedError(
                    "Automated selection of `botorch_acqf_class` is not supported "
                    "for robust optimization with risk measures. Please specify "
                    "`botorch_acqf_class` as part of `model_kwargs`."
                )
            self._botorch_acqf_class = choose_botorch_acqf_class(
                torch_opt_config=torch_opt_config,
                datasets=self.surrogate.training_data,
                use_p_feasible=self.use_p_feasible,
            )
        return self.acquisition_class(
            surrogate=self.surrogate,
            botorch_acqf_class=self._botorch_acqf_class,
            botorch_acqf_options=botorch_acqf_options,
            botorch_acqf_classes_with_options=botorch_acqf_classes_with_options,
            search_space_digest=search_space_digest,
            torch_opt_config=torch_opt_config,
            options=acq_options,
            n=n,
        )

    def feature_importances(self) -> npt.NDArray:
        """Compute feature importances from the model.

        This assumes that we can get model lengthscales from either
        ``covar_module.base_kernel.lengthscale`` or ``covar_module.lengthscale``.

        Returns:
            The feature importances as a numpy array of size len(metrics) x 1 x dim
            where each row sums to 1.
        """
        return get_feature_importances_from_botorch_model(model=self.surrogate.model)

    @property
    def search_space_digest(self) -> SearchSpaceDigest:
        """Returns the ``SearchSpaceDigest`` that was used while fitting the underlying
        surrogate. If the surrogate has not been fit, raises a ``ModelError``.
        """
        if self.surrogate._last_search_space_digest is None:
            raise ModelError(
                "`search_space_digest` is not initialized. Must `fit` the model first."
            )
        return self.surrogate._last_search_space_digest



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Functions that create ``BenchmarkProblem``s based on BoTorch test functions.
"""

from collections.abc import Mapping
from typing import Any, Literal

from ax.benchmark.benchmark_problem import (
    BenchmarkProblem,
    get_continuous_search_space,
    get_moo_opt_config,
    get_soo_opt_config,
)
from ax.benchmark.benchmark_step_runtime_function import TBenchmarkStepRuntimeFunction
from ax.benchmark.benchmark_test_functions.botorch_test import BoTorchTestFunction
from ax.core.auxiliary import AuxiliaryExperiment, AuxiliaryExperimentPurpose
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.core.types import TParameterization
from botorch.test_functions.base import (
    BaseTestProblem,
    ConstrainedBaseTestProblem,
    MultiObjectiveTestProblem,
)
from botorch.test_functions.multi_fidelity import AugmentedBranin

# A mapping from (BoTorch problem class name, dim | None) to baseline value
# Obtained using `get_baseline_value_from_sobol`
BOTORCH_BASELINE_VALUES: Mapping[tuple[str, int | None], float] = {
    ("Ackley", 4): 19.837273921447853,
    ("Branin", None): 10.930455126654936,
    ("BraninCurrin", None): 0.9820209831769217,
    ("BraninCurrin", 30): 3.0187520516793587,
    ("ConstrainedGramacy", None): 1.0643958597443999,
    ("ConstrainedBraninCurrin", None): 0.9820209831769217,
    ("Griewank", 4): 60.037068040081095,
    ("Hartmann", 3): -2.3423173903286716,
    ("Hartmann", 6): -0.796988050854654,
    ("Hartmann", 30): -0.8359462084890045,
    ("KeaneBumpFunction", 2): -0.0799243632005311,
    ("KeaneBumpFunction", 10): -0.13609325522288143,
    ("Levy", 4): 14.198811442165178,
    ("Powell", 4): 932.3102865964689,
    ("PressureVessel", None): float("inf"),
    ("Rosenbrock", 4): 30143.767857949348,
    ("SixHumpCamel", None): 0.45755007063109004,
    ("SpeedReducer", None): float("inf"),
    ("TensionCompressionString", None): float("inf"),
    ("ThreeHumpCamel", None): 3.7321680621434155,
    ("WeldedBeamSO", None): float("inf"),
}


def _get_name(
    test_problem: BaseTestProblem,
    observe_noise_sd: bool,
    dim: int | None = None,
    n_dummy_dimensions: int = 0,
) -> str:
    """
    Get a string name describing the problem, in a format such as
    "hartmann_fixed_noise_6d" or "jenatton" (where the latter would
    not have fixed noise and have the default dimensionality).
    """
    base_name = f"{test_problem.__class__.__name__}"
    observed_noise = "_observed_noise" if observe_noise_sd else ""
    if dim is None and n_dummy_dimensions == 0:
        dim_str = ""
    else:
        total_dim = (test_problem.dim if dim is None else dim) + n_dummy_dimensions
        dim_str = f"_{total_dim}d"
    return f"{base_name}{observed_noise}{dim_str}"


def create_problem_from_botorch(
    *,
    test_problem_class: type[BaseTestProblem],
    test_problem_kwargs: dict[str, Any],
    noise_std: float | list[float] = 0.0,
    num_trials: int,
    baseline_value: float | None = None,
    name: str | None = None,
    lower_is_better: bool = True,
    observe_noise_sd: bool = False,
    use_shifted_function: bool = False,
    search_space: SearchSpace | None = None,
    report_inference_value_as_trace: bool = False,
    step_runtime_function: TBenchmarkStepRuntimeFunction | None = None,
    status_quo_params: TParameterization | None = None,
    auxiliary_experiments_by_purpose: (
        dict[AuxiliaryExperimentPurpose, list[AuxiliaryExperiment]] | None
    ) = None,
    n_dummy_dimensions: int = 0,
    use_map_metric: bool = False,
    n_steps: int = 1,
) -> BenchmarkProblem:
    """
    Create a ``BenchmarkProblem`` from a BoTorch ``BaseTestProblem``.

    The resulting ``BenchmarkProblem``'s ``test_function`` is constructed from
    the ``BaseTestProblem`` class (``test_problem_class``) and its arguments
    (``test_problem_kwargs``). All other fields are passed to
    ``BenchmarkProblem`` if they are specified and populated with reasonable
    defaults otherwise. ``num_trials``, however, must be specified.

    Args:
        test_problem_class: The BoTorch test problem class which will be used
            to define the `search_space`, `optimization_config`, and `runner`.
        test_problem_kwargs: Keyword arguments used to instantiate the
            `test_problem_class`. This should *not* include `noise_std` or
            `negate`, since these are handled through Ax benchmarking (as the
            `noise_std` and `lower_is_better` arguments to `BenchmarkProblem`).
        noise_std: Standard deviation of synthetic noise added to outcomes. If a
            float, the same noise level is used for all objectives.
        lower_is_better: Whether this is a minimization problem. For MOO, this
            applies to all objectives.
        num_trials: Simply the `num_trials` of the `BenchmarkProblem` created.
        baseline_value: If not provided, will be looked up from
            `BOTORCH_BASELINE_VALUES`.
        name: Will be passed to ``BenchmarkProblem`` if specified and populated
            with reasonable defaults otherwise.
        observe_noise_sd: Whether the standard deviation of the observation noise is
            observed or not (in which case it must be inferred by the model).
            This is separate from whether synthetic noise is added to the
            problem, which is controlled by the `noise_std` of the test problem.
        use_shifted_function: Whether to use the shifted version of the test function.
            If True, an offset tensor is randomly drawn from the test problem bounds,
            and the we evaluate `f(X-offset)` rather than `f(X)`. This is useful for
            changing the location of the optima for test functions that favor the
            center of the search space.
            If True, the default search space is enlarged to include the optimizer
            within the bounds even after the offset is applied. If the original
            bounds were `[low, high]`, the new bounds are `[2*low, 2*high]`.
        search_space: If provided, the `search_space` of the `BenchmarkProblem`.
            Otherwise, a `SearchSpace` with all `RangeParameter`s is created
            from the bounds of the test problem.
        report_inference_value_as_trace: If True, indicates that the
            ``optimization_trace`` on a ``BenchmarkResult`` ought to be the
            ``inference_trace``; otherwise, it will be the ``oracle_trace``.
            See ``BenchmarkResult`` for more information.
        status_quo_params: The status quo parameters for the problem.
        step_runtime_function: Optionally, a function that takes in ``params``
            (typically dictionaries mapping strings to ``TParamValue``s) and
            returns the runtime of an step. If ``step_runtime_function`` is
            left as ``None``, each step will take one simulated second.  (When
            data is not time-series, the whole trial consists of one step.)
        auxiliary_experiments_by_purpose: A mapping from experiment purpose to
            a list of auxiliary experiments.
        n_dummy_dimensions: If >0, the search space will be augmented
            with extra dimensions. The corresponding parameters will have no
            effect on function values.
        use_map_metric: Whether to use a ``BenchmarkMapMetric`` (rather than a
            ``BenchmarkMetric``).
        n_steps: Number of steps (progression values) in each evaluation. The
            default of 1 reflects a normal synthetic function evaluation. A
            higher number results in repeating the evaluation and getting the
            same result ``n_steps`` times (before IID noise is added).

    Example:
        >>> from ax.benchmark.benchmark_problem import create_problem_from_botorch
        >>> from botorch.test_functions.synthetic import Branin
        >>> problem = create_problem_from_botorch(
        ...    test_problem_class=Branin,
        ...    test_problem_kwargs={},
        ...    noise_std=0.1,
        ...    num_trials=10,
        ...    observe_noise_sd=True,
        ...    step_runtime_function=lambda params: 1 / params["fidelity"],
        ... )
    """
    # pyre-fixme [45]: Invalid class instantiation
    test_problem = test_problem_class(**test_problem_kwargs)
    is_constrained = isinstance(test_problem, ConstrainedBaseTestProblem)

    if search_space is None:
        if use_shifted_function:
            search_space = get_continuous_search_space(
                bounds=[(2 * b[0], 2 * b[1]) for b in test_problem._bounds],
                n_dummy_dimensions=n_dummy_dimensions,
            )
        else:
            search_space = get_continuous_search_space(
                bounds=test_problem._bounds, n_dummy_dimensions=n_dummy_dimensions
            )

    dim = test_problem_kwargs.get("dim", None)

    n_obj = test_problem.num_objectives
    if name is None:
        name = _get_name(
            test_problem=test_problem,
            observe_noise_sd=observe_noise_sd,
            dim=dim,
            n_dummy_dimensions=n_dummy_dimensions,
        )

    num_constraints = test_problem.num_constraints if is_constrained else 0
    if isinstance(test_problem, MultiObjectiveTestProblem):
        # pyre-fixme[6]: For 1st argument expected `SupportsIndex` but got
        #  `Union[Tensor, Module]`.
        objective_names = [f"{name}_{i}" for i in range(n_obj)]
    else:
        objective_names = [name]

    # pyre-fixme[6]: For 1st argument expected `SupportsIndex` but got `Union[int,
    #  Tensor, Module]`.
    constraint_names = [f"constraint_slack_{i}" for i in range(num_constraints)]
    outcome_names = objective_names + constraint_names

    test_function = BoTorchTestFunction(
        botorch_problem=test_problem,
        outcome_names=outcome_names,
        use_shifted_function=use_shifted_function,
        dummy_param_names={
            n for n in search_space.parameters if "embedding_dummy_" in n
        },
        n_steps=n_steps,
    )

    if isinstance(test_problem, MultiObjectiveTestProblem):
        optimization_config = get_moo_opt_config(
            # pyre-fixme[6]: For 1st argument expected `int` but got `Union[int,
            #  Tensor, Module]`.
            num_constraints=num_constraints,
            lower_is_better=lower_is_better,
            observe_noise_sd=observe_noise_sd,
            outcome_names=test_function.outcome_names,
            ref_point=test_problem._ref_point,
            use_map_metric=use_map_metric,
        )
    else:
        optimization_config = get_soo_opt_config(
            outcome_names=test_function.outcome_names,
            lower_is_better=lower_is_better,
            observe_noise_sd=observe_noise_sd,
            use_map_metric=use_map_metric,
        )

    optimal_value = (
        test_problem.max_hv
        if isinstance(test_problem, MultiObjectiveTestProblem)
        else test_problem.optimal_value
    )
    baseline_value = (
        BOTORCH_BASELINE_VALUES[(test_problem_class.__name__, dim)]
        if baseline_value is None
        else baseline_value
    )

    return BenchmarkProblem(
        name=name,
        search_space=search_space,
        optimization_config=optimization_config,
        test_function=test_function,
        noise_std=noise_std,
        num_trials=num_trials,
        # pyre-fixme[6]: For 7th argument expected `float` but got `Union[float,
        #  Tensor, Module]`.
        optimal_value=optimal_value,
        baseline_value=baseline_value,
        report_inference_value_as_trace=report_inference_value_as_trace,
        step_runtime_function=step_runtime_function,
        status_quo_params=status_quo_params,
        auxiliary_experiments_by_purpose=auxiliary_experiments_by_purpose,
    )


def get_augmented_branin_search_space(
    fidelity_or_task: Literal["fidelity", "task"],
) -> SearchSpace:
    """
    Get the ``SearchSpace`` that matches the ``AugmentedBranin`` test problem.

    ``AugmentedBranin`` has an extra parameter beyond the normal two which has
    been treated as a fidelity parameter.

    Args:
        fidelity_or_task: If "fidelity", the extra parameter is a fidelity
            parameter and will be continuous, because fidelity ChoiceParameters
            can't be used with the ``OrderedChoiceToIntegerRange`` transform. If
            "task", the extra parameter is a task parameter and is discrete,
            because a ``RangeParameter`` cannot be a task.
    """
    if fidelity_or_task == "fidelity":
        extra_parameter = RangeParameter(
            name="x2",
            parameter_type=ParameterType.FLOAT,
            lower=0.0,
            upper=1.0,
            is_fidelity=True,
            target_value=1,
        )
    else:
        extra_parameter = ChoiceParameter(
            name="x2",
            parameter_type=ParameterType.FLOAT,
            values=[0, 1],
            is_fidelity=False,
            is_task=True,
            target_value=1,
        )
    parameters = [
        RangeParameter(
            name=f"x{i}",
            parameter_type=ParameterType.FLOAT,
            lower=0.0,
            upper=1.0,
        )
        for i in range(2)
    ] + [extra_parameter]
    return SearchSpace(parameters=parameters)


def get_augmented_branin_problem(
    fidelity_or_task: Literal["fidelity", "task"],
    report_inference_value_as_trace: bool = True,
) -> BenchmarkProblem:
    """
    Get a Branin problem with a fidelity or task parameter.

    Args:
        fidelity_or_task: If "fidelity", the extra parameter is a fidelity
            parameter. If "task", the extra parameter is a task parameter.
        report_inference_value_as_trace: Passed to
            ``create_problem_from_botorch`` then to ``BenchmarkProblem``.
    """

    return create_problem_from_botorch(
        test_problem_class=AugmentedBranin,
        test_problem_kwargs={},
        search_space=get_augmented_branin_search_space(
            fidelity_or_task=fidelity_or_task
        ),
        num_trials=3,
        baseline_value=3.0,
        report_inference_value_as_trace=report_inference_value_as_trace,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.utils.testing.backend_simulator import BackendSimulator, BackendSimulatorOptions


def get_backend_simulator_with_trials() -> BackendSimulator:
    options = BackendSimulatorOptions(
        internal_clock=0.0, use_update_as_start_time=True, max_concurrency=2
    )
    sim = BackendSimulator(options=options)
    sim.run_trial(0, 2)
    sim.run_trial(1, 1)
    sim.run_trial(2, 10)
    return sim



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.core.arm import Arm
from ax.core.experiment import Experiment
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import SearchSpace
from ax.exceptions.generation_strategy import (
    AxGenerationException,
    GenerationStrategyRepeatedPoints,
)
from ax.generation_strategy.center_generation_node import CenterGenerationNode
from ax.generation_strategy.generation_node import logger
from ax.generation_strategy.transition_criterion import AutoTransitionAfterGen
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from pyre_extensions import none_throws


class TestCenterGenerationNode(TestCase):
    def test_center_generation(self) -> None:
        ss = SearchSpace(
            parameters=[
                RangeParameter(  # Simple float.
                    name="x1",
                    parameter_type=ParameterType.FLOAT,
                    lower=-5.0,
                    upper=10.0,
                ),
                RangeParameter(  # Integer with log transform.
                    name="x2",
                    parameter_type=ParameterType.INT,
                    lower=10.0,
                    upper=100.0,
                    log_scale=True,
                ),
                ChoiceParameter(  # Ordered choice.
                    name="x3",
                    parameter_type=ParameterType.STRING,
                    values=["a", "b", "c", "d"],
                    is_ordered=True,
                ),
                FixedParameter(  # Fixed parameter.
                    name="x4",
                    parameter_type=ParameterType.BOOL,
                    value=True,
                ),
            ]
        )
        node = CenterGenerationNode(next_node_name="test")
        self.assertEqual(node.next_node_name, "test")
        self.assertEqual(
            node.transition_criteria,
            [
                AutoTransitionAfterGen(
                    transition_to="test", continue_trial_generation=False
                )
            ],
        )
        experiment = Experiment(search_space=ss)
        params = (
            none_throws(node.gen(experiment=experiment, pending_observations=None))
            .arms[0]
            .parameters
        )
        self.assertEqual(node.search_space, ss)
        self.assertEqual(params, {"x1": 2.5, "x2": 31, "x3": "c", "x4": True})

    def test_deduplication(self) -> None:
        exp = get_branin_experiment()
        exp.new_trial().add_arm(arm=Arm({"x1": 2.5, "x2": 7.5})).run()
        node = CenterGenerationNode(next_node_name="test")
        gr = none_throws(node.gen(experiment=exp, pending_observations=None))
        self.assertEqual(gr._model_key, "Sobol")

    def test_repr(self) -> None:
        node = CenterGenerationNode(next_node_name="test")
        self.assertEqual(
            repr(node),
            "CenterGenerationNode(next_node_name='test')",
        )

    def test_equality(self) -> None:
        node = CenterGenerationNode(next_node_name="test")
        node2 = CenterGenerationNode(next_node_name="test")
        self.assertEqual(node, node2)
        other_node = CenterGenerationNode(next_node_name="test2")
        self.assertNotEqual(node, other_node)
        # Still equal after generation, despite the search spaces being different.
        # The two nodes will function the same if we call gen with a different
        # experiment after this, so I think this is fine.
        exp = get_branin_experiment()
        node.gen(experiment=exp, pending_observations=None)
        self.assertEqual(node, node2)

    def test_with_constraints(self) -> None:
        ss = SearchSpace(
            parameters=[
                RangeParameter(
                    name="x1",
                    parameter_type=ParameterType.FLOAT,
                    lower=-5.0,
                    upper=10.0,
                ),
                RangeParameter(
                    name="x2",
                    parameter_type=ParameterType.INT,
                    lower=10.0,
                    upper=100.0,
                    log_scale=True,
                ),
            ],
            parameter_constraints=[  # x1 <= 0
                ParameterConstraint(constraint_dict={"x1": 1.0}, bound=0.0)
            ],
        )
        node = CenterGenerationNode(next_node_name="test")
        self.assertEqual(
            set(node.fallback_specs.keys()),
            {AxGenerationException, GenerationStrategyRepeatedPoints},
        )
        # Generate and check for Sobol fallback.
        exp = Experiment(search_space=ss)
        with self.assertLogs(logger=logger) as logs:
            gr = none_throws(node.gen(experiment=exp, pending_observations=None))
        self.assertTrue(
            any(
                "Center of the search space does not satisfy parameter constraints."
                in log.message
                for log in logs.records
            )
        )
        self.assertEqual(gr._model_key, "Sobol")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

# NOTE: Do not add `from __future__ import annotations` to this file. Adding
# `annotations` postpones evaluation of types and will break FBLearner's usage of
# `BenchmarkResult` as return type annotation, used for serialization and rendering
# in the UI.

from collections.abc import Iterable
from dataclasses import dataclass

import numpy.typing as npt
from ax.core.experiment import Experiment
from ax.utils.common.base import Base
from numpy import nanmean, nanquantile
from pandas import DataFrame
from scipy.stats import sem

PERCENTILES = [0.25, 0.5, 0.75]


@dataclass(eq=False)
class BenchmarkResult(Base):
    """The result of a single optimization loop from one
    (BenchmarkProblem, BenchmarkMethod) pair.

    Args:
        name: Name of the benchmark. Should make it possible to determine the
            problem and the method.
        seed: Seed used for determinism.
        oracle_trace: For single-objective problems, the oracle trace is the
            best oracle objective value seen on completed trials up to that
            point. For multi-objective problems, it is the cumulative
            hypervolume of feasible oracle objective values.

            Oracle values are typically objective values that are at the ground
            truth (not noisy) and evaluated at the target task and fidelity.

            The trace may have fewer elements than the number of trials run if
            multiple trials stop at the same time; the trace is updated whenever
            trials stop (TrialStatus COMPLETED or EARLY_STOPPED). The number of
            trials completed is reflected in the `cost_trace`, which is updated
            at the same time as the `oracle_trace`. For example, if each trial
            has a cost of 1, and `cost_trace[i] = 4`, then `oracle_trace[i]` is
            the value of the best of the first four trials to complete, or the
            feasible hypervolume of those trials.
        inference_trace: Inference values come from choosing a "best" point or
            points based only on data that would be observable in realistic
            settings, as specified by `BenchmarkMethod.get_best_parameters`, and
            then evaluating the oracle objective value of that point according
            to the problem's `OptimizationConfig`.

            By default, if it is not overridden,
            `BenchmarkMethod.get_best_parameters` uses the empirical best point
            if `use_model_predictions_for_best_point` is False and the best
            point of those evaluated so far if it is True.

            As with the oracle trace, the inference trace is updated whenever a
            trial completes or stops early and may have fewer elements than the
            number of trials of multiple trials complete at the same time.

            Note: This is scaled differently from "inference regret", which is a
            lower-is-better value that is relative to the best possible value.
            The inference value trace is higher-is-better if the problem is a
            maximization problem or if the problem is multi-objective (in which
            case hypervolume is used). Hence, it is signed the same as
            `oracle_trace` and `optimization_trace`. `score_trace`, meanwhile,
            is higher-is-better and relative to the optimum.
        optimization_trace: Either the `oracle_trace` or the `inference_trace`,
            depending on whether the `BenchmarkProblem` specifies
            `report_inference_value`. Having `optimization_trace` specified
            separately is useful when we need just one value to evaluate how
            well the benchmark went.
        score_trace: The scores associated with the problem, typically either
            the optimization_trace or inference_value_trace normalized to a
            0-100 scale for comparability between problems.
        cost_trace: The cumulative cost of completed trials. The `cost_trace` is
            updated whenever a trial completes, so, like the `oracle_trace` and
            `inference_trace`, it can have fewer elements than the number of
            trials if multiple trials complete at the same time. Trials that do
            not produce `MapData` have a cost of 1, and trials that produce
            `MapData` have a cost equal to the length of the `MapData`.
        fit_time: Total time spent fitting models.
        gen_time: Total time spent generating candidates.
        experiment: If not ``None``, the Ax experiment associated with the
            optimization that generated this data. Either ``experiment`` or
            ``experiment_storage_id`` must be provided.
        experiment_storage_id: Pointer to location where experiment data can be read.
    """

    name: str
    seed: int

    oracle_trace: npt.NDArray
    inference_trace: npt.NDArray
    optimization_trace: npt.NDArray
    score_trace: npt.NDArray
    cost_trace: npt.NDArray

    fit_time: float
    gen_time: float

    experiment: Experiment | None = None
    experiment_storage_id: str | None = None

    def __post_init__(self) -> None:
        if self.experiment is not None and self.experiment_storage_id is not None:
            raise ValueError(
                "Cannot specify both an `experiment` and an "
                "`experiment_storage_id` for the experiment."
            )
        if self.experiment is None and self.experiment_storage_id is None:
            raise ValueError(
                "Must provide an `experiment` or `experiment_storage_id` "
                "to construct a BenchmarkResult."
            )


@dataclass(frozen=True, eq=False)
class AggregatedBenchmarkResult(Base):
    """The result of a benchmark test, or series of replications. Scalar data present
    in the BenchmarkResult is here represented as (mean, sem) pairs.
    """

    name: str
    results: list[BenchmarkResult]

    # mean, sem, and quartile columns
    optimization_trace: DataFrame
    score_trace: DataFrame

    # (mean, sem) pairs
    fit_time: list[float]
    gen_time: list[float]

    @classmethod
    def from_benchmark_results(
        cls,
        results: list[BenchmarkResult],
    ) -> "AggregatedBenchmarkResult":
        """Aggregrates a list of BenchmarkResults. For various reasons (timeout, errors,
        etc.) each BenchmarkResult may have a different number of trials; aggregated
        traces and statistics are computed with and truncated to the minimum trial count
        to ensure each replication is included.
        """
        # Extract average wall times and standard errors thereof
        fit_time, gen_time = (
            [nanmean(Ts), float(sem(Ts, ddof=1, nan_policy="propagate"))]
            for Ts in zip(*((res.fit_time, res.gen_time) for res in results))
        )

        # Compute some statistics for each trace
        trace_stats = {}
        for name in ("optimization_trace", "score_trace"):
            step_data = zip(*(getattr(res, name) for res in results))
            stats = _get_stats(step_data=step_data, percentiles=PERCENTILES)
            trace_stats[name] = stats

        # Return aggregated results
        return cls(
            name=results[0].name,
            results=results,
            fit_time=fit_time,
            gen_time=gen_time,
            **{name: DataFrame(stats) for name, stats in trace_stats.items()},
        )


def _get_stats(
    step_data: Iterable[npt.NDArray],
    percentiles: list[float],
) -> dict[str, list[float]]:
    quantiles = []
    stats = {"mean": [], "sem": []}
    for step_vals in step_data:
        stats["mean"].append(nanmean(step_vals))
        stats["sem"].append(sem(step_vals, ddof=1, nan_policy="propagate"))
        quantiles.append(nanquantile(step_vals, q=percentiles))
    stats.update({f"P{100 * p:.0f}": q for p, q in zip(percentiles, zip(*quantiles))})
    return stats



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
import pandas as pd
from ax.analysis.summary import Summary
from ax.api.client import Client
from ax.api.configs import RangeParameterConfig
from ax.core.base_trial import TrialStatus
from ax.core.trial import Trial
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_offline_experiments, get_online_experiments
from pyre_extensions import assert_is_instance, none_throws


class TestSummary(TestCase):
    def test_compute(self) -> None:
        client = Client()
        client.configure_experiment(
            name="test_experiment",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
                RangeParameterConfig(
                    name="x2",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
            ],
        )
        client.configure_optimization(objective="foo, bar")

        # Get two trials and fail one, giving us a ragged structure
        client.get_next_trials(max_trials=2)
        client.complete_trial(trial_index=0, raw_data={"foo": 1.0, "bar": 2.0})
        client.mark_trial_failed(trial_index=1)

        analysis = Summary()

        with self.assertRaisesRegex(UserInputError, "requires an `Experiment`"):
            analysis.compute()

        experiment = client._experiment
        card = analysis.compute(experiment=experiment)

        # Test metadata
        self.assertEqual(card.name, "Summary")
        self.assertEqual(card.title, "Summary for test_experiment")
        self.assertEqual(
            card.subtitle,
            "High-level summary of the `Trial`-s in this `Experiment`",
        )
        self.assertIsNotNone(card.blob)

        # Test dataframe for accuracy
        self.assertEqual(
            {*card.df.columns},
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "foo",
                "bar",
                "x1",
                "x2",
            },
        )

        trial_0_parameters = none_throws(
            assert_is_instance(experiment.trials[0], Trial).arm
        ).parameters
        trial_1_parameters = none_throws(
            assert_is_instance(experiment.trials[1], Trial).arm
        ).parameters
        expected = pd.DataFrame(
            {
                "trial_index": {0: 0, 1: 1},
                "arm_name": {0: "0_0", 1: "1_0"},
                "trial_status": {0: "COMPLETED", 1: "FAILED"},
                "generation_node": {0: "CenterOfSearchSpace", 1: "Sobol"},
                "foo": {0: 1.0, 1: np.nan},  # NaN because trial 1 failed
                "bar": {0: 2.0, 1: np.nan},
                "x1": {
                    0: trial_0_parameters["x1"],
                    1: trial_1_parameters["x1"],
                },
                "x2": {
                    0: trial_0_parameters["x2"],
                    1: trial_1_parameters["x2"],
                },
            }
        )
        pd.testing.assert_frame_equal(card.df, expected)

        # Test without omitting empty columns
        analysis_no_omit = Summary(omit_empty_columns=False)
        card_no_omit = analysis_no_omit.compute(experiment=experiment)
        self.assertEqual(
            {*card_no_omit.df.columns},
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "fail_reason",
                "generation_node",
                "foo",
                "bar",
                "x1",
                "x2",
            },
        )
        self.assertEqual(len(card_no_omit.df), len(experiment.arms_by_name))

    def test_online(self) -> None:
        # Test MetricSummary can be computed for a variety of experiments which
        # resemble those we see in an online setting.

        for omit_empty_columns in [True, False]:
            analysis = Summary(omit_empty_columns=omit_empty_columns)

            for experiment in get_online_experiments():
                _ = analysis.compute(experiment=experiment)

    def test_offline(self) -> None:
        # Test MetricSummary can be computed for a variety of experiments which
        # resemble those we see in an offline setting.

        for omit_empty_columns in [True, False]:
            analysis = Summary(omit_empty_columns=omit_empty_columns)

            for experiment in get_offline_experiments():
                _ = analysis.compute(experiment=experiment)

    def test_trial_indices_filter(self) -> None:
        """Test that Client.summarize correctly uses Summary."""
        client = Client()
        client.configure_experiment(
            name="test_experiment",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
                RangeParameterConfig(
                    name="x2",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
            ],
        )
        client.configure_optimization(objective="foo")

        # Get a trial
        client.get_next_trials(max_trials=1)
        client.complete_trial(trial_index=0, raw_data={"foo": 1.0})

        # Test summarize with trial_indices
        df_filtered = client.summarize(trial_indices=[0])
        self.assertEqual(len(df_filtered), 1)

        # Test that changes to the experiment are reflected in the summary
        client.get_next_trials(max_trials=1)
        client.complete_trial(trial_index=1, raw_data={"foo": 2.0})

    def test_trial_status_filter(self) -> None:
        """Test that Summary correctly filters by trial_status."""
        client = Client()
        client.configure_experiment(
            name="test_experiment",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
            ],
        )
        client.configure_optimization(objective="foo")

        # Create trials with different statuses
        client.get_next_trials(max_trials=1)
        client.complete_trial(trial_index=0, raw_data={"foo": 1.0})

        client.get_next_trials(max_trials=1)
        client.mark_trial_failed(trial_index=1)

        client.get_next_trials(max_trials=1)
        # Trial 2 remains in RUNNING state

        # Test filtering by completed status
        analysis = Summary(trial_statuses=[TrialStatus.COMPLETED])
        experiment = client._experiment
        card = analysis.compute(experiment=experiment)
        self.assertEqual(len(card.df), 1)
        self.assertEqual(card.df["trial_index"].iloc[0], 0)
        self.assertEqual(card.df["trial_status"].iloc[0], "COMPLETED")

        # Test filtering by failed status
        analysis = Summary(trial_statuses=[TrialStatus.FAILED])
        card = analysis.compute(experiment=experiment)
        self.assertEqual(len(card.df), 1)
        self.assertEqual(card.df["trial_index"].iloc[0], 1)
        self.assertEqual(card.df["trial_status"].iloc[0], "FAILED")

        # Test filtering by running status
        analysis = Summary(trial_statuses=[TrialStatus.RUNNING])
        card = analysis.compute(experiment=experiment)
        self.assertEqual(len(card.df), 1)
        self.assertEqual(card.df["trial_index"].iloc[0], 2)
        self.assertEqual(card.df["trial_status"].iloc[0], "RUNNING")

        # Test filtering by multiple statuses
        analysis = Summary(trial_statuses=[TrialStatus.COMPLETED, TrialStatus.FAILED])
        card = analysis.compute(experiment=experiment)
        self.assertEqual(len(card.df), 2)
        self.assertIn(0, card.df["trial_index"].values)
        self.assertIn(1, card.df["trial_index"].values)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Sequence

import pandas as pd
from ax.utils.common.base import SortableBase
from ax.utils.tutorials.environment import is_running_in_papermill
from IPython.display import display, HTML, Markdown

from plotly.offline import get_plotlyjs

# Simple HTML template for rendering a card with a title, subtitle, and body with
# scrollable overflow.
html_card_template = """
<style>
.card {{
    overflow: auto;
    border-width: thin;
    border-color: lightgray;
    border-style: solid;
    border-radius: 0.5em;
    padding: 10px;
}}
.card-header:hover {{
    cursor: pointer;
}}
</style>
<div class="card">
    <div class="card-header">
        <details>
            <summary><b>{title_str}</b></summary>
            <p>{subtitle_str}</p>
        </details>
    </div>
    <div class="card-body">
        {body_html}
    </div>
</div>
"""

# Simple HTML template for rendering a *group* card with a title, subtitle, and
# body with scrollable overflow.
html_group_card_template = """
<style>
.group-card {{
    overflow: auto;
    margin-top: 25px;
}}
.group-header {{
    font-size: 1.5em;
}}
.group-subtitle {{
    font-size: 1em;
}}
</style>
<div class="group-card">
    <div>
        <b class="group-header">
        {title_str}
        </b>
        <p class="group-subtitle">
        {subtitle_str}
        </p>
    </div>
    <div class="group-body">
        {body_html}
    </div>
</div>
"""

# HTML template for putting cards into a 2 x N CSS grid.
html_grid_template = """
<style>
    .grid-container {{
        display: grid;
        grid-template-columns: repeat(2, 2fr);
        gap: 10px;
    }}
</style>
<div class="grid-container">
    {card_divs}
</div>
"""


class AnalysisCardBase(SortableBase, ABC):
    """
    Abstract base class for "cards", the result of a call to Analyis.compute(...).
    Cards may either be a single card (AnalysisCard and its subclasses) or an ordered
    collection of cards (AnalysisCardGroup) -- together these three classes form a
    tree structure which can hold arbitrarily nested collections of cards.

    When rendering in an IPython environment (ex. Jupyter), use AnalyisCardBase.flatten
    to produce an ordered list of cards to render.
    """

    name: str

    title: str
    subtitle: str

    _timestamp: datetime

    def __init__(
        self,
        name: str,
        title: str,
        subtitle: str,
        timestamp: datetime | None = None,
    ) -> None:
        """
        Args:
            name: The class name of the Analysis that produced this card (ex. "Summary",
                "ArmEffects", etc.).
            title: Human-readable title which describes the card's contents. This
                appears in all user facing interfaces as the title of the card.
            subtitle: Human-readable subtitle which provides additional information or
                context to improve the usability of the analysis for users.
            timestamp: The time at which the Analysis was computed. This can be
                especially useful when querying the database for the most recently
                produced artifacts.
        """
        self.name = name
        self.title = title
        self.subtitle = subtitle
        self._timestamp = timestamp if timestamp is not None else datetime.now()

    @abstractmethod
    def flatten(self) -> list[AnalysisCard]:
        """
        Returns a list of AnalysisCards contained in this card in order. This is useful
        when processing a collection of cards where order is necessary but grouping can
        safely be ignored (ex. when rendering a collection of cards in an IPython).
        """
        pass

    @abstractmethod
    def hierarchy_str(self, level: int = 0) -> str:
        """
        Returns a string representation of the card's nested hierarchy structure. This
        is useful for debugging and logging.

        Example:
            Root
                Child 1
                    Grandchild 1
                    Grandchild 2
                Child 2
                    Grandchild 3
        """
        pass

    @property
    def _unique_id(self) -> str:
        return str(hash(str(self.__dict__)))

    def _ipython_display_(self) -> None:
        """
        IPython display hook. This is called when the AnalysisCard is rendered in an
        IPython environment (ex. Jupyter). This method should not be implemented by
        subclasses; instead they should implement the representation-specific helpers
        such as _body_html_ and _body_papermill_.
        """

        # If in papermill used simplified rendering. This is used for the Ax website.
        if is_running_in_papermill():
            for card in self.flatten():
                display(Markdown(f"**{card.title}**\n\n{card.subtitle}"))
                display(card._body_papermill())

            return

        display(HTML(self._repr_html_()))

    @abstractmethod
    def _body_html(self, depth: int) -> str:
        """
        Return the HTML body of the card (the dataframe, plot, grid, etc.). This is
        used by the AnalysisCardBase._repr_html_ method to render the card in an
        IPython environment (ex. Jupyter).

        This, not _repr_html_ or _to_html, should be implemented by subclasses of
        AnalysisCardBase in most cases in order to keep treatment of titles and
        subtitles consistent.

        Since this method can sometimes be called recursively a "depth" parameter can
        be passed in as well.
        """
        pass

    def _repr_html_(self) -> str:
        """
        IPython HTML representation hook. This is called when the AnalysisCard is
        rendered in an IPython environment (ex. Jupyter). This method should be
        implemented by subclasses of Analysis to display the AnalysisCard in a useful
        way.
        """
        # require.js is not compatible with ES6 import used by plotly.js so we must
        # null out `define` here. This does not affect rendering.
        plotlyjs_script = f"<script>define = null;{get_plotlyjs()}</script>"

        return plotlyjs_script + self._to_html(depth=0)

    def _to_html(self, depth: int) -> str:
        return html_card_template.format(
            title_str=self.title,
            subtitle_str=self.subtitle,
            body_html=self._body_html(depth=depth),
        )


class AnalysisCardGroup(AnalysisCardBase):
    """
    An ordered collection of AnalysisCards. This is useful for grouping related
    analyses together.

    This is analogous to a "branch node" in a tree structure.
    """

    children: list[AnalysisCardBase]

    def __init__(
        self,
        name: str,
        title: str,
        subtitle: str | None,
        children: Sequence[AnalysisCardBase],
        timestamp: datetime | None = None,
    ) -> None:
        """
        Args:
            name: The class name of the Analysis that produced this card (ex. "Summary",
                "ArmEffects", etc.).
            title: Human-readable title which describes the card's contents. This
                appears in all user facing interfaces as the title of the card.
            subtitle: Human-readable subtitle which provides additional information or
                context to improve the usability of the analysis for users.
            timestamp: The time at which the Analysis was computed. This can be
                especially useful when querying the database for the most recently
                produced artifacts.
        """
        super().__init__(
            name=name,
            title=title,
            subtitle=subtitle if subtitle is not None else "",
            timestamp=timestamp,
        )

        self.children = [
            child
            for child in children
            # Filter out empty analysis card groups
            if not (isinstance(child, AnalysisCardGroup) and len(child.children) == 0)
        ]

    @property
    def _unique_id(self) -> str:
        return self.name

    def flatten(self) -> list[AnalysisCard]:
        return [child for child in self.children for child in child.flatten()]

    def hierarchy_str(self, level: int = 0) -> str:
        return f"{'    ' * level}{self.name}\n" + "\n".join(
            child.hierarchy_str(level=level + 1) for child in self.children
        )

    def _body_html(self, depth: int) -> str:
        """
        When rendering an AnalysisCardGroup as HTML use the following rules when
        constructing the card's body:

        * Render children in order
        * Render AnalysisCards (leaves) in a 2xN grid when adjacent to each other
        * Do not render AnalysisCardGroups in a grid
        * Do not render subtitles below depth == 2 (this is handled in
            AnalysisCardBase._to_html, not this method).
        """

        res = []
        leaf_cards = []
        for child in self.children:
            # Accumulate adjacent AnalysisCard leaves so they can be inserted into an
            # HTML grid later.
            if isinstance(child, AnalysisCard):
                leaf_cards.append(child)
                continue

            # If there are leaves accumulated, collect them into a grid and empty
            # the accumulator before appending the current child AnalysisCardGroup
            # underneath.
            if len(leaf_cards) > 0:
                leaves_grid = html_grid_template.format(
                    card_divs="".join(
                        [card._to_html(depth=depth + 1) for card in leaf_cards]
                    )
                )
                res.append(leaves_grid)
                leaf_cards = []

            res.append(child._to_html(depth=depth + 1))

        # Collect the accumulated leaves a final time to append to the result.
        if len(leaf_cards) > 0:
            leaves_grid = html_grid_template.format(
                card_divs="".join(
                    [card._to_html(depth=depth + 1) for card in leaf_cards]
                )
            )
            res.append(leaves_grid)

        return "\n".join(res)

    def _to_html(self, depth: int) -> str:
        return html_group_card_template.format(
            title_str=self.title,
            subtitle_str=self.subtitle,
            body_html=self._body_html(depth=depth),
        )


class AnalysisCard(AnalysisCardBase):
    """
    The ultimate result of a call to Analysis.compute(...). This holds the raw data
    produced by the compute function as a dataframe, and some arbitrary blob of data
    which will be rendered in the card in a notebook or a UI front-end (e.g. a Plotly
    figure, Markdown formatted text etc.)

    Subclasses of AnalysisCard define the structure of the blob (ex. a Plotly Figure)
    and implement methods for rendering the card in a useful way.

    This is analogous to a "leaf node" in a tree structure.
    """

    df: pd.DataFrame  # Raw data produced by the Analysis

    # Blob is the data processed for end-user consumption, encoded as a string,
    # typically JSON. Subclasses of Analysis can define their own methods for consuming
    # the blob and presenting it to the user (ex. PlotlyAnalysisCard.get_figure()
    # decodes the blob into a go.Figure object).
    blob: str

    def __init__(
        self,
        name: str,
        title: str,
        subtitle: str,
        df: pd.DataFrame,
        blob: str,
        timestamp: datetime | None = None,
    ) -> None:
        """
        Args:
            name: The name of the Analysis that produced this card.
            title: A human-readable title which describes the card's contents.
            subtitle: A human-readable subtitle which elaborates on the card's title if
                necessary.
            df: The raw data produced by the Analysis.
            blob: The data processed for end-user consumption, encoded as a string,
                typically JSON.
            timestamp: The time at which the Analysis was computed. This can be
                especially useful when querying the database for the most recently
                produced artifacts.
        """
        super().__init__(
            name=name,
            title=title,
            subtitle=subtitle,
            timestamp=timestamp,
        )

        self.df = df
        self.blob = blob

    @property
    def _unique_id(self) -> str:
        return self.title

    def flatten(self) -> list[AnalysisCard]:
        return [self]

    def hierarchy_str(self, level: int = 0) -> str:
        return f"{'    ' * level}{self.title}"

    def _body_html(self, depth: int) -> str:
        """
        By default, this method displays the raw data in a pandas DataFrame.
        """

        return f"<div class='content'>{self.df.to_html()}</div>"

    def _body_papermill(self) -> Any:
        """
        Return the body of the AnalysisCard in a simplified format for when html is
        undesirable (ex. when rendering the Ax website).

        By default, this method displays the raw data in a pandas DataFrame.
        """

        return self.df


class ErrorAnalysisCard(AnalysisCard):
    # TODO: Implement improved rendering which shows the traceback.
    # def _ipython_display_(self) -> None: ...
    pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from logging import WARNING

import numpy as np
from ax.adapter.registry import Generators
from ax.benchmark.benchmark import benchmark_replication
from ax.benchmark.methods.modular_botorch import get_sobol_botorch_modular_acquisition
from ax.benchmark.methods.sobol import get_sobol_benchmark_method
from ax.benchmark.problems.registry import get_benchmark_problem
from ax.utils.common.testutils import TestCase
from ax.utils.testing.mock import mock_botorch_optimize
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.analytic import LogExpectedImprovement
from botorch.acquisition.knowledge_gradient import qKnowledgeGradient
from botorch.acquisition.logei import qLogExpectedImprovement
from botorch.models.gp_regression import SingleTaskGP
from pyre_extensions import none_throws


class TestMethods(TestCase):
    def _test_mbm_acquisition(self, batch_size: int) -> None:
        method = get_sobol_botorch_modular_acquisition(
            model_cls=SingleTaskGP,
            acquisition_cls=qKnowledgeGradient,
            batch_size=batch_size,
            distribute_replications=False,
        )
        is_batched = batch_size > 1
        expected_name = "MBM::SingleTaskGP_qKnowledgeGradient" + (
            f"_q{batch_size}" if is_batched else ""
        )
        self.assertEqual(method.name, expected_name)
        gs = method.generation_strategy
        sobol, kg = gs._steps
        self.assertEqual(kg.generator, Generators.BOTORCH_MODULAR)
        model_kwargs = none_throws(kg.model_kwargs)
        self.assertEqual(model_kwargs["botorch_acqf_class"], qKnowledgeGradient)
        surrogate_spec = model_kwargs["surrogate_spec"]
        self.assertEqual(
            surrogate_spec.model_configs[0].botorch_model_class.__name__,
            "SingleTaskGP",
        )

    def test_mbm_acquisition(self) -> None:
        for batch_size in [1, 2]:
            with self.subTest(batch_size=batch_size):
                self._test_mbm_acquisition(batch_size=batch_size)

    @mock_botorch_optimize
    def _test_benchmark_replication_runs(
        self, batch_size: int, acqf_cls: type[AcquisitionFunction]
    ) -> None:
        problem = get_benchmark_problem(problem_key="ackley4")
        method = get_sobol_botorch_modular_acquisition(
            model_cls=SingleTaskGP,
            batch_size=batch_size,
            acquisition_cls=acqf_cls,
            num_sobol_trials=2,
            name="test",
            distribute_replications=False,
        )
        n_sobol_trials = method.generation_strategy._steps[0].num_trials
        self.assertEqual(n_sobol_trials, 2)
        self.assertEqual(method.name, "test")
        # Only run one non-Sobol trial
        n_total_trials = n_sobol_trials + 1
        problem = get_benchmark_problem(
            problem_key="ackley4", num_trials=n_total_trials
        )
        result = benchmark_replication(
            problem=problem, method=method, seed=0, orchestrator_logging_level=WARNING
        )
        self.assertTrue(np.isfinite(result.score_trace).all())
        self.assertEqual(result.optimization_trace.shape, (n_total_trials,))

        self.assertEqual(
            len(none_throws(result.experiment).arms_by_name),
            n_total_trials * batch_size,
        )

    def test_benchmark_replication_runs(self) -> None:
        with self.subTest(name="sequential LogEI"):
            self._test_benchmark_replication_runs(
                batch_size=1,
                acqf_cls=LogExpectedImprovement,
            )
        with self.subTest(name="sequential qLogEI"):
            self._test_benchmark_replication_runs(
                batch_size=1,
                acqf_cls=qLogExpectedImprovement,
            )
        with self.subTest(name="batch qLogEI"):
            self._test_benchmark_replication_runs(
                batch_size=2,
                acqf_cls=qLogExpectedImprovement,
            )

    def test_sobol(self) -> None:
        method = get_sobol_benchmark_method(distribute_replications=False)
        self.assertEqual(method.name, "Sobol")
        gs = method.generation_strategy
        self.assertEqual(len(gs._steps), 1)
        self.assertEqual(gs._steps[0].generator, Generators.SOBOL)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
from ax.adapter.transforms.log_y import match_ci_width
from ax.utils.common.testutils import TestCase
from numpy import typing as npt


class TestUtils(TestCase):
    def test_match_ci_width(self) -> None:
        # Tests the behavior of match_ci_width using a simple transform.

        def transform(y: npt.NDArray) -> npt.NDArray:
            return y * 2.0

        mean = np.array([1.0, 2.0])
        variance = np.array([0.1, 0.2])
        # Transform using variance.
        new_mean, new_variance = match_ci_width(
            mean=mean,
            sem=None,
            variance=variance,
            transform=transform,
        )
        self.assertTrue(np.allclose(new_mean, np.array([2.0, 4.0])))
        self.assertTrue(np.allclose(new_variance, np.array([0.4, 0.8])))
        # Transform using sem.
        new_mean, new_sem = match_ci_width(
            mean=mean,
            sem=np.sqrt(variance),
            variance=None,
            transform=transform,
        )
        self.assertTrue(np.allclose(new_mean, np.array([2.0, 4.0])))
        self.assertTrue(np.allclose(new_sem, np.array([np.sqrt(0.4), np.sqrt(0.8)])))
        # Transform with NaN variance.
        new_mean, new_variance = match_ci_width(
            mean=mean,
            sem=None,
            variance=np.array([np.nan, np.nan]),
            transform=transform,
        )
        self.assertTrue(np.allclose(new_mean, np.array([2.0, 4.0])))
        self.assertTrue(np.all(np.isnan(new_variance)))
        # Test with upper and lower bounds.
        new_mean, new_sem = match_ci_width(
            mean=mean,
            sem=np.array([0.1, 0.2]),
            variance=None,
            transform=transform,
            lower_bound=1.0,
            upper_bound=1.9,
        )
        self.assertTrue(np.allclose(new_mean, np.array([2.0, 3.8])))
        # Bounds are set to clip the CI to the mean from one side only.
        # We end up with doubling the halved CI width, ending up with the original sem.
        self.assertTrue(np.allclose(new_sem, np.array([0.1, 0.2])))
